#!/usr/bin/env python

"""
Unit test of comms for libensemble.
"""

import time
import queue
import logging
from libensemble.tools.tools import osx_set_mp_method

import numpy as np
import libensemble.comms.comms as comms
import libensemble.comms.logs as commlogs


osx_set_mp_method()


def test_qcomm():
    "Test queue-based bidirectional communicator."

    inq = queue.Queue()
    outq = queue.Queue()
    comm = comms.QComm(inq, outq)

    comm.send('a', 1)
    comm.send('b')
    assert outq.get() == ('a', 1) and outq.get() == ('b',) and outq.empty(), \
        "Check send appropriately goes to output queue."

    comm.push_to_buffer('b', 0)
    inq.put(('c', 3))
    inq.put(('d',))
    assert (comm.recv() == ('b', 0)
            and comm.recv() == ('c', 3)
            and comm.recv() == ('d',)
            and inq.empty()), \
        "Check recv appropriately comes from input queue."

    flag = True
    try:
        comm.recv(0.1)
        flag = False
    except comms.Timeout:
        pass
    assert flag, "Check comms receive returns Timeout at appropriate time."


def test_missing_handler():
    "Test correct ABC complaint about missing message handler"

    flag = True
    try:
        class TestHandler(comms.GenCommHandler):
            "Dummy GenCommHandler"

            def on_worker_avail(self, nworker):
                return "on_worker_avail", nworker

            def on_queued(self, sim_id):
                return "on_queued", sim_id

            # Missing on_result

            def on_update(self, sim_id, recs):
                return "on_update", sim_id, recs

            def on_killed(self, sim_id):
                return "on_killed", sim_id

        TestHandler(None)
        flag = False

    except TypeError:
        pass

    assert flag, "Check define-time error on missing handlers"


def test_gen_comm_handler():
    "Test GenCommHandler interface"

    class TestGenComm(comms.GenCommHandler):
        "Dummy GenComm handler"

        def on_worker_avail(self, nworker):
            return "on_worker_avail", nworker

        def on_queued(self, sim_id):
            return "on_queued", sim_id

        def on_result(self, sim_id, recs):
            return "on_result", sim_id, recs

        def on_update(self, sim_id, recs):
            return "on_update", sim_id, recs

        def on_killed(self, sim_id):
            return "on_killed", sim_id

    inq = queue.Queue()
    outq = queue.Queue()
    comm = comms.QComm(inq, outq)
    gcomm = TestGenComm(comm)

    gcomm.send_request(None)
    gcomm.send_kill(10)
    gcomm.send_get_history(1, 2)
    gcomm.send_subscribe()

    assert outq.get() == ('request', None)
    assert outq.get() == ('kill', 10)
    assert outq.get() == ('get_history', 1, 2)
    assert outq.get() == ('subscribe',)
    assert outq.empty()

    inq.put(('worker_avail', 3))
    inq.put(('queued', 1))
    inq.put(('result', 1, 100))
    inq.put(('update', 1, 50))
    inq.put(('killed', 1))
    inq.put(('qwerty',))

    assert gcomm.process_message() == ('on_worker_avail', 3)
    assert gcomm.process_message() == ('on_queued', 1)
    assert gcomm.process_message() == ('on_result', 1, 100)
    assert gcomm.process_message() == ('on_update', 1, 50)
    assert gcomm.process_message() == ('on_killed', 1)

    flag = True
    try:
        gcomm.process_message(0.1)
    except ValueError:
        pass
    assert flag, "Check error handling for bad message type"

    flag = True
    try:
        gcomm.process_message(0.1)
        flag = False
    except comms.Timeout:
        pass
    assert flag, "Check timeout on process_message"

    flag = True
    try:
        inq.put(('stop',))
        gcomm.process_message(0.1)
        flag = False
    except comms.ManagerStop:
        pass
    assert flag, "Check exception raised on manager requested stop"


def test_sim_comm_handler():
    "Test SimCommHandler interface"

    class TestSimComm(comms.SimCommHandler):
        "Dummy SimCommHandler"

        def on_request(self, sim_id, recs):
            return "on_request", sim_id, recs

        def on_kill(self, sim_id):
            return "on_kill", sim_id

    inq = queue.Queue()
    outq = queue.Queue()
    comm = comms.QComm(inq, outq)
    scomm = TestSimComm(comm)

    scomm.send_result(1, None)
    scomm.send_update(1, 10)
    scomm.send_killed(1)

    assert outq.get() == ('result', 1, None)
    assert outq.get() == ('update', 1, 10)
    assert outq.get() == ('killed', 1)
    assert outq.empty()

    inq.put(('request', 1, 100))
    inq.put(('kill', 1))
    inq.put(('qwerty',))

    assert scomm.process_message() == ('on_request', 1, 100)
    assert scomm.process_message() == ('on_kill', 1)

    flag = True
    try:
        scomm.process_message(0.1)
    except ValueError:
        pass
    assert flag, "Check error handling for bad message type"

    flag = True
    try:
        scomm.process_message(0.1)
        flag = False
    except comms.Timeout:
        pass
    assert flag, "Check timeout on process_message"

    flag = True
    try:
        inq.put(('stop',))
        scomm.process_message(0.1)
        flag = False
    except comms.ManagerStop:
        pass
    assert flag, "Check exception raised on manager requested stop"


def test_comm_eval():
    "Test CommEval and Future interfaces"

    gen_specs = {'out': [('x', float), ('flag', bool)]}

    inq = queue.Queue()
    outq = queue.Queue()
    comm = comms.QComm(inq, outq)
    gcomm = comms.CommEval(comm, gen_specs=gen_specs)

    inq.put(('worker_avail', 3))
    inq.put(('queued', 1))
    H_o = np.zeros(2, dtype=gen_specs['out'])
    promises = gcomm.request(H_o)

    assert len(promises) == 2
    assert gcomm.workers == 3
    assert gcomm.sim_started == 2
    assert gcomm.sim_pending == 2
    assert outq.get(timeout=0.1)[0] == 'request'
    assert outq.empty()

    inq.put(('queued', 3))
    promise = gcomm(10.0, True)
    assert outq.get(timeout=0.1)[0] == 'request'
    assert not promise.cancelled() and not promise.done()
    assert gcomm.sim_started == 3
    assert gcomm.sim_pending == 3

    flag = True
    try:
        promise.result(timeout=0.1)
        flag = False
    except comms.Timeout:
        pass
    assert flag, "Checking timeout on result check"

    promise.cancel()
    assert outq.get(timeout=0.1) == ('kill', 3)
    assert not promise.cancelled() and not promise.done()
    inq.put(('killed', 3))
    gcomm.process_message(timeout=0.1)
    assert promise.cancelled() and promise.done()
    assert promise.result() is None
    assert gcomm.sim_started == 3
    assert gcomm.sim_pending == 2

    results = np.zeros(3, dtype=[('f', float)])
    results['f'] = [20, 10, 15]
    resultsf = results['f']

    resultsf[0] = 15
    inq.put(('update', 1, results[0]))
    gcomm.process_message(timeout=0.1)
    assert not promises[0].cancelled() and not promises[0].done()
    assert promises[0].current_result == 15
    assert gcomm.sim_started == 3
    assert gcomm.sim_pending == 2

    resultsf[0] = 20
    inq.put(('result', 1, results[0]))
    gcomm.process_message(timeout=0.1)
    assert not promises[0].cancelled() and promises[0].done()
    assert promises[0].result() == 20
    assert gcomm.sim_started == 3
    assert gcomm.sim_pending == 1

    inq.put(('update', 2, results[1]))
    gcomm.process_message(timeout=0.1)
    assert not promises[1].cancelled() and not promises[1].done()
    assert gcomm.sim_started == 3
    assert gcomm.sim_pending == 1

    inq.put(('result', 2, results[2]))
    gcomm.process_message(timeout=0.1)
    assert not promises[1].cancelled() and promises[1].done()
    assert gcomm.sim_started == 3
    assert gcomm.sim_pending == 0

    inq.put(('queued', 4))
    promise = gcomm(x=5.0)
    msg_type, recs = outq.get()
    assert msg_type == 'request'
    assert recs['x'][0] == 5.0 and not recs['flag'][0]
    assert gcomm.sim_started == 4 and gcomm.sim_pending == 1

    inq.put(('killed', 4))
    gcomm.process_message(timeout=0.1)
    assert promise.cancelled() and promise.done()
    assert gcomm.sim_started == 4 and gcomm.sim_pending == 0


def run_qcomm_threadproc_test(ThreadProc):
    "Test CommEval between two threads or processes (allows timeout checks)"

    class BadWorkerException(Exception):
        pass

    def worker_thread(comm, gen_specs):
        gcomm = comms.CommEval(comm, gen_specs=gen_specs)
        p1 = gcomm(x=0.5)
        p2 = gcomm(x=1.0)
        x1 = p1.result()
        assert x1 == 5
        assert not p2.done()
        gcomm.wait_all()
        assert p2.done()
        assert p2.result(timeout=0) == 10
        p3 = gcomm(x=3)
        p4 = gcomm(x=4)
        assert not p3.done()
        assert not p4.done()
        gcomm.wait_any()
        assert p3.done()
        return 128

    def bad_worker_thread(comm):
        raise BadWorkerException("Bad worker")

    gen_specs = {'out': [('x', float), ('flag', bool)]}
    results = np.zeros(3, dtype=[('f', float)])
    results['f'] = [5, 10, 30]
    # resultsf = results['f']
    with ThreadProc(worker_thread, gen_specs=gen_specs) as mgr_comm:
        assert mgr_comm.running
        assert mgr_comm.recv()[0] == 'request'
        mgr_comm.send('queued', 0)
        assert mgr_comm.recv()[0] == 'request'
        mgr_comm.send('queued', 1)
        time.sleep(0.2)
        assert not mgr_comm.mail_flag()
        mgr_comm.send('result', 0, results[0])
        time.sleep(0.5)
        mgr_comm.send('result', 1, results[1])
        mgr_comm.send('queued', 2)
        mgr_comm.send('queued', 3)
        time.sleep(0.5)
        mgr_comm.send('result', 2, results[2])
        assert mgr_comm.result() == 128

    try:
        bad_worker_okay = True
        with ThreadProc(bad_worker_thread) as comm:

            flag = True
            try:
                comm.recv(0.1)
                flag = False
            except comms.Timeout:
                pass
            except comms.CommFinishedException:
                pass
            assert flag, "Test receive timeout from worker"

            _ = comm.result()
            bad_worker_okay = False
    except BadWorkerException:
        pass
    except comms.RemoteException as e:
        assert str(e) == "Bad worker"

    assert bad_worker_okay, "Checking bad worker flag"


def test_qcomm_threadproc():
    "Test CommEval between threads and processes"
    run_qcomm_threadproc_test(comms.QCommThread)
    run_qcomm_threadproc_test(comms.QCommProcess)


def test_comm_logging():
    "Test logging from a worker process is handled correctly."

    def worker_main(comm):
        ch = commlogs.CommLogHandler(comm)
        logger = logging.getLogger()
        logger.addHandler(ch)
        logger.setLevel(logging.INFO)
        logger.info("Test message")
        comm.send("Done!")

    with comms.QCommProcess(worker_main) as mgr_comm:
        msg = mgr_comm.recv()
        assert isinstance(msg[0], logging.LogRecord)



"""
Runs libEnsemble testing the MPI Runners command creation.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_mpi_runners.py
   python3 test_mpi_runners.py --nworkers 3 --comms local
   python3 test_mpi_runners.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

import numpy as np
from libensemble.libE import libE
from libensemble.sim_funcs.run_line_check import runline_check as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor
from libensemble.tests.regression_tests.common import create_node_file
from libensemble import logger

# logger.set_level('DEBUG')  # For testing the test
logger.set_level('INFO')

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 2 4

nworkers, is_manager, libE_specs, _ = parse_args()
rounds = 1
sim_app = '/path/to/fakeapp.x'
comms = libE_specs['comms']

libE_specs['dedicated_mode'] = True
libE_specs['enforce_worker_core_bounds'] = True

# To allow visual checking - log file not used in test
log_file = 'ensemble_mpi_runners_comms_' + str(comms) + '_wrks_' + str(nworkers) + '.log'
logger.set_filename(log_file)

nodes_per_worker = 2

# For varying size test - relate node count to nworkers
node_file = 'nodelist_mpi_runners_comms_' + str(comms) + '_wrks_' + str(nworkers)
nnodes = nworkers * nodes_per_worker

if is_manager:
    create_node_file(num_nodes=nnodes, name=node_file)

if comms == 'mpi':
    libE_specs['mpi_comm'].Barrier()

# Mock up system
custom_resources = {
    'cores_on_node': (16, 64),  # Tuple (physical cores, logical cores)
    'node_file': node_file,  # Name of file containing a node-list
}
libE_specs['resource_info'] = custom_resources

persis_info = add_unique_random_streams({}, nworkers + 1)
exit_criteria = {'sim_max': nworkers * rounds}

# TODO: May move specs, inputs and expected outputs to a data_set module.
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, (2,))],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'gen_batch_size': 100,
    },
}

# Each worker has 2 nodes. Basic test list for portable options
test_list_base = [
    {'testid': 'base1', 'nprocs': 2, 'nnodes': 1, 'ppn': 2, 'e_args': '--xarg 1'},  # Under use
    {'testid': 'base2'},  # Give no config and no extra_args
    {'testid': 'base3', 'e_args': '--xarg 1'},  # Give no config with extra_args
    {'testid': 'base4', 'e_args': '--xarg 1', 'ht': True},  # Give no config but with HT
    {'testid': 'base5', 'nprocs': 16, 'e_args': '--xarg 1'},  # Just nprocs (will use one node)
    {'testid': 'base6', 'nprocs': 16, 'nnodes': 2, 'e_args': '--xarg 1'},  # nprocs, nnodes
]

# extra_args tests for each mpi runner
# extra_args should be parsed - however, the string should stay intact, inc abbreviated/different expressions
eargs_mpich = [
    {'testid': 'mp1', 'nprocs': 16, 'e_args': '--xarg 1 --ppn 16'},  # nprocs + parse extra_args
    {'testid': 'mp2', 'e_args': '-np 8 --xarg 1 --ppn 4'},  # parse extra_args
]

eargs_openmpi = [
    {'testid': 'ompi1', 'nprocs': 16, 'e_args': '--xarg 1 -npernode 16'},  # nprocs + parse extra_args
    {'testid': 'ompi2', 'e_args': '-np 8 --xarg 1 -npernode 4'},  # parse extra_args
]

eargs_aprun = [
    {'testid': 'ap1', 'nprocs': 16, 'e_args': '--xarg 1 -N 16'},  # nprocs + parse extra_args
    {'testid': 'ap2', 'e_args': '-n 8 --xarg 1 -N 4'},  # parse extra_args
]

# Note in a8: -n 8 is abbreviated form of --ntasks, this should be unaltered while --nodes is derived and inserted.
eargs_srun = [
    {'testid': 'sr1', 'nprocs': 16, 'e_args': '--xarg 1 --ntasks-per-node 16'},  # nprocs + parse extra_args
    {'testid': 'sr2', 'e_args': '-n 8 --xarg 1 --ntasks-per-node 4'},  # parse extra_args
]

# Note for jsrun: proc = resource set. Awkward naming but this seems like the best solution.
# Define extra_args as minimal relation of tasks/cores/gpus (one resource set), then n (nprocs) as multiplier.
eargs_jsrun = [
    {'testid': 'jsr1', 'nprocs': 16, 'e_args': '--xarg 1 -r 16'},  # nprocs + parse extra_args
    {'testid': 'jsr2', 'e_args': '-n 8 --xarg 1 -r 4'},  # parse extra_args
    {'testid': 'jsr3', 'nprocs': 3, 'e_args': '-a 1 -c 1 -g 1 --bind=packed:1 --smpiargs="-gpu"'},  # combine r-sets
    {'testid': 'jsr4', 'e_args': '-n 3 -a 1 -c 1 -g 1 --bind=packed:1 --smpiargs="-gpu"'},  # r-sets all in extra_args
]

eargs_custom = [{'testid': 'cust1', 'e_args': '--xarg 1 --ppn 16'}]

# Expected Outputs
# Node IDs are modified for different workers in sim func.
exp_mpich = [
    'mpirun -hosts node-1 -np 2 --ppn 2 --xarg 1 /path/to/fakeapp.x --testid base1',
    'mpirun -hosts node-1,node-2 -np 32 --ppn 16 /path/to/fakeapp.x --testid base2',
    'mpirun -hosts node-1,node-2 -np 32 --ppn 16 --xarg 1 /path/to/fakeapp.x --testid base3',
    'mpirun -hosts node-1,node-2 -np 128 --ppn 64 --xarg 1 /path/to/fakeapp.x --testid base4',
    'mpirun -hosts node-1 -np 16 --ppn 16 --xarg 1 /path/to/fakeapp.x --testid base5',
    'mpirun -hosts node-1,node-2 -np 16 --ppn 8 --xarg 1 /path/to/fakeapp.x --testid base6',
    'mpirun -hosts node-1 -np 16 --xarg 1 --ppn 16 /path/to/fakeapp.x --testid mp1',
    'mpirun -hosts node-1,node-2 -np 8 --xarg 1 --ppn 4 /path/to/fakeapp.x --testid mp2',
]

exp_rename_mpich = [
    'inst -dummy mpich -hosts node-1 -np 2 --ppn 2 --xarg 1 /path/to/fakeapp.x --testid base1',
    'inst -dummy mpich -hosts node-1,node-2 -np 32 --ppn 16 /path/to/fakeapp.x --testid base2',
    'inst -dummy mpich -hosts node-1,node-2 -np 32 --ppn 16 --xarg 1 /path/to/fakeapp.x --testid base3',
    'inst -dummy mpich -hosts node-1,node-2 -np 128 --ppn 64 --xarg 1 /path/to/fakeapp.x --testid base4',
    'inst -dummy mpich -hosts node-1 -np 16 --ppn 16 --xarg 1 /path/to/fakeapp.x --testid base5',
    'inst -dummy mpich -hosts node-1,node-2 -np 16 --ppn 8 --xarg 1 /path/to/fakeapp.x --testid base6',
    'inst -dummy mpich -hosts node-1 -np 16 --xarg 1 --ppn 16 /path/to/fakeapp.x --testid mp1',
    'inst -dummy mpich -hosts node-1,node-2 -np 8 --xarg 1 --ppn 4 /path/to/fakeapp.x --testid mp2',
]

# openmpi requires machinefiles (-host requires
# exp_openmpi = \
#    ['mpirun -host node-1 -np 2 -npernode 2 --xarg 1 /path/to/fakeapp.x --testid base1',
#    'mpirun -host node-1,node-2 -np 32 -npernode 16 /path/to/fakeapp.x --testid base2',
#    'mpirun -host node-1,node-2 -np 32 -npernode 16 --xarg 1 /path/to/fakeapp.x --testid base3',
#    'mpirun -host node-1,node-2 -np 128 -npernode 64 --xarg 1 /path/to/fakeapp.x --testid base4',
#    'mpirun -host node-1 -np 16 -npernode 16 --xarg 1 /path/to/fakeapp.x --testid base5',
#    'mpirun -host node-1,node-2 -np 16 -npernode 8 --xarg 1 /path/to/fakeapp.x --testid base6',
#    'mpirun -host node-1 -np 16 --xarg 1 -npernode 16 /path/to/fakeapp.x --testid ompi1',
#    'mpirun -host node-1,node-2 -np 8 --xarg 1 -npernode 4 /path/to/fakeapp.x --testid ompi2',
#    ]

exp_aprun = [
    'aprun -L node-1 -n 2 -N 2 --xarg 1 /path/to/fakeapp.x --testid base1',
    'aprun -L node-1,node-2 -n 32 -N 16 /path/to/fakeapp.x --testid base2',
    'aprun -L node-1,node-2 -n 32 -N 16 --xarg 1 /path/to/fakeapp.x --testid base3',
    'aprun -L node-1,node-2 -n 128 -N 64 --xarg 1 /path/to/fakeapp.x --testid base4',
    'aprun -L node-1 -n 16 -N 16 --xarg 1 /path/to/fakeapp.x --testid base5',
    'aprun -L node-1,node-2 -n 16 -N 8 --xarg 1 /path/to/fakeapp.x --testid base6',
    'aprun -L node-1 -n 16 --xarg 1 -N 16 /path/to/fakeapp.x --testid ap1',
    'aprun -L node-1,node-2 -n 8 --xarg 1 -N 4 /path/to/fakeapp.x --testid ap2',
]

exp_srun = [
    'srun -w node-1 --ntasks 2 --nodes 1 --ntasks-per-node 2 --xarg 1 /path/to/fakeapp.x --testid base1',
    'srun -w node-1,node-2 --ntasks 32 --nodes 2 --ntasks-per-node 16 /path/to/fakeapp.x --testid base2',
    'srun -w node-1,node-2 --ntasks 32 --nodes 2 --ntasks-per-node 16 --xarg 1 /path/to/fakeapp.x --testid base3',
    'srun -w node-1,node-2 --ntasks 128 --nodes 2 --ntasks-per-node 64 --xarg 1 /path/to/fakeapp.x --testid base4',
    'srun -w node-1 --ntasks 16 --nodes 1 --ntasks-per-node 16 --xarg 1 /path/to/fakeapp.x --testid base5',
    'srun -w node-1,node-2 --ntasks 16 --nodes 2 --ntasks-per-node 8 --xarg 1 /path/to/fakeapp.x --testid base6',
    'srun -w node-1 --ntasks 16 --nodes 1 --xarg 1 --ntasks-per-node 16 /path/to/fakeapp.x --testid sr1',
    'srun -w node-1,node-2 --nodes 2 -n 8 --xarg 1 --ntasks-per-node 4 /path/to/fakeapp.x --testid sr2',
]

exp_jsrun = [
    'jsrun -n 2 -r 2 --xarg 1 /path/to/fakeapp.x --testid base1',
    'jsrun -n 32 /path/to/fakeapp.x --testid base2',
    'jsrun -n 32 --xarg 1 /path/to/fakeapp.x --testid base3',
    'jsrun -n 128 --xarg 1 /path/to/fakeapp.x --testid base4',
    'jsrun -n 16 --xarg 1 /path/to/fakeapp.x --testid base5',
    'jsrun -n 16 -r 8 --xarg 1 /path/to/fakeapp.x --testid base6',
    'jsrun -n 16 --xarg 1 -r 16 /path/to/fakeapp.x --testid jsr1',
    'jsrun -n 8 --xarg 1 -r 4 /path/to/fakeapp.x --testid jsr2',
    'jsrun -n 3 -a 1 -c 1 -g 1 --bind=packed:1 --smpiargs="-gpu" /path/to/fakeapp.x --testid jsr3',
    'jsrun -n 3 -a 1 -c 1 -g 1 --bind=packed:1 --smpiargs="-gpu" /path/to/fakeapp.x --testid jsr4',
]

exp_custom = [
    'myrunner --xarg 1 /path/to/fakeapp.x --testid base1',
    'myrunner /path/to/fakeapp.x --testid base2',
    'myrunner --xarg 1 /path/to/fakeapp.x --testid base3',
    'myrunner --xarg 1 /path/to/fakeapp.x --testid base4',
    'myrunner --xarg 1 /path/to/fakeapp.x --testid base5',
    'myrunner --xarg 1 /path/to/fakeapp.x --testid base6',
    'myrunner --xarg 1 --ppn 16 /path/to/fakeapp.x --testid cust1',
]


# Loop here for mocking different systems.
def run_tests(mpi_runner, runner_name, test_list_exargs, exp_list):

    mpi_customizer = {
        'mpi_runner': mpi_runner,  # Select runner: mpich, openmpi, aprun, srun, jsrun
        'runner_name': runner_name,  # Runner name: Replaces run command if not None
    }

    exctr = MPIExecutor(custom_info=mpi_customizer)
    exctr.register_app(full_path=sim_app, calc_type='sim')

    test_list = test_list_base + test_list_exargs
    sim_specs['user'] = {
        'tests': test_list,
        'expect': exp_list,
        'nodes_per_worker': nodes_per_worker,
        'persis_gens': 0,
    }

    # Perform the run
    H, pinfo, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)


# for run_set in ['mpich', 'openmpi', 'aprun', 'srun', 'jsrun', 'rename_mpich', 'custom']:
for run_set in ['mpich', 'aprun', 'srun', 'jsrun', 'rename_mpich', 'custom']:

    # Could use classes, pref in separate data_set module
    runner_name = None  # Use default

    if run_set == 'mpich':
        runner = 'mpich'
        test_list_exargs = eargs_mpich
        exp_list = exp_mpich

    #    if run_set == 'openmpi':
    #        runner = 'openmpi'
    #        test_list_exargs = eargs_openmpi
    #        exp_list = exp_openmpi

    if run_set == 'aprun':
        runner = 'aprun'
        test_list_exargs = eargs_aprun
        exp_list = exp_aprun

    if run_set == 'srun':
        runner = 'srun'
        test_list_exargs = eargs_srun
        exp_list = exp_srun

    if run_set == 'jsrun':
        runner = 'jsrun'
        test_list_exargs = eargs_jsrun
        exp_list = exp_jsrun

    if run_set == 'rename_mpich':
        runner = 'mpich'
        runner_name = 'inst -dummy mpich'
        test_list_exargs = eargs_mpich
        exp_list = exp_rename_mpich

    # Auto-resources will not parse - everything goes in extra_args
    if run_set == 'custom':
        runner = 'custom'
        runner_name = 'myrunner'
        test_list_exargs = eargs_custom
        exp_list = exp_custom

    run_tests(runner, runner_name, test_list_exargs, exp_list)

# All asserts are in sim func



import numpy as np
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def give_sim_work_first(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function gives (in order) entries in ``H`` to idle workers
    to evaluate in the simulation function. The fields in ``sim_specs['in']``
    are given. If all entries in `H` have been given to be evaluated, a worker
    is told to call the generator function, provided this wouldn't result in
    more than ``alloc_specs['user']['num_active_gen']`` active generators. Also allows
    for a ``'batch_mode'``.

    .. seealso::
        `test_old_aposmm_with_gradients.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_old_aposmm_with_gradients.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    Work = {}
    gen_count = support.count_gens()

    for wid in support.avail_worker_ids():
        # Skip any cancelled points
        while persis_info['next_to_give'] < len(H) and H[persis_info['next_to_give']]['cancel_requested']:
            persis_info['next_to_give'] += 1

        # Find indices of H that are not yet allocated
        if persis_info['next_to_give'] < len(H):
            # Give sim work if possible
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], [persis_info['next_to_give']], [])
            except InsufficientFreeResources:
                break
            persis_info['next_to_give'] += 1

        elif gen_count < user.get('num_active_gens', gen_count + 1):
            lw = persis_info['last_worker']

            last_size = persis_info.get('last_size')

            if len(H):
                # Don't give gen instances in batch mode if points are unfinished
                if user.get('batch_mode') and not support.all_returned(pt_filter=~H['paused'], low_bound=last_size):
                    break
                # Don't call APOSMM if there are runs going but none need advancing
                if len(persis_info[lw]['run_order']):
                    runs_needing_to_advance = np.zeros(len(persis_info[lw]['run_order']), dtype=bool)
                    for run, inds in enumerate(persis_info[lw]['run_order'].values()):
                        runs_needing_to_advance[run] = H['returned'][inds[-1]]

                    if not np.any(runs_needing_to_advance):
                        break

            # Give gen work
            try:
                Work[wid] = support.gen_work(wid, gen_specs['in'], range(len(H)), persis_info[lw])
            except InsufficientFreeResources:
                break

            gen_count += 1
            persis_info['total_gen_calls'] += 1
            persis_info['last_worker'] = wid
            persis_info['last_size'] = len(H)

    return Work, persis_info



"""
@Author: Created by Caleb Ju at Argonne National Labs as Given's associate,
         Summer 2021.
@About: Based on n-agent with gradient tracking:
        https://ieeexplore.ieee.org/abstract/document/9199106/
"""
import numpy as np
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG
from libensemble.tools.consensus_subroutines import print_final_score, get_grad, get_neighbor_vals


def n_agent(H, persis_info, gen_specs, libE_info):
    """ Gradient sliding. Coordinates with alloc to do local and distributed
        (i.e., gradient of consensus step) calculations.
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(ub)

    # start with random x0
    x0 = persis_info['rand_stream'].uniform(low=lb, high=ub, size=(n,))
    x_k = x0

    L = persis_info['params']['L']
    eps = persis_info['params']['eps']
    rho = persis_info['params']['rho']
    N_const = persis_info['params']['N_const']
    step_const = persis_info['params']['step_const']

    N = int(N_const/eps + 1)
    eta = step_const * 1.0/L * min(1/6, (1-rho**2)**2/(4*rho**2*(3+4*rho**2)))

    f_i_idxs = persis_info['f_i_idxs']
    A_i_data = persis_info['A_i_data']
    A_i_gen_ids = persis_info['A_i_gen_ids']
    local_gen_id = persis_info['worker_num']

    # sort A_i to be increasing gen_id
    _perm_ids = np.argsort(A_i_gen_ids)
    A_weights = A_i_data[_perm_ids]
    A_i_gen_ids = A_i_gen_ids[_perm_ids]
    A_i_gen_ids_no_local = np.delete(A_i_gen_ids,
                                     np.where(A_i_gen_ids == local_gen_id)[0][0])

    prev_s_is = np.zeros((len(A_i_data), n), dtype=float)
    prev_gradf_is = np.zeros((len(A_i_data), n), dtype=float)

    if local_gen_id == 1:
        print('[{}%]: '.format(0), flush=True, end='')
    print_final_score(x_k, f_i_idxs, gen_specs, libE_info)
    percent = 0.1

    for k in range(N):
        tag, gradf = get_grad(x_k, f_i_idxs, gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        tag, neighbor_gradf_is = get_neighbor_vals(gradf, local_gen_id,
                                                   A_i_gen_ids_no_local,
                                                   gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        U = (prev_s_is + neighbor_gradf_is - prev_gradf_is)
        # takes linear combination as described by equation (9)
        s = np.dot(U.T, A_weights)

        tag, neighbor_x_is = get_neighbor_vals(x_k, local_gen_id,
                                               A_i_gen_ids_no_local,
                                               gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG
        tag, neighbor_s_is = get_neighbor_vals(s, local_gen_id,
                                               A_i_gen_ids_no_local,
                                               gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        V = (neighbor_x_is - eta * neighbor_s_is)
        # takes linear combination as described by equation (10)
        next_x_k = np.dot(V.T, A_weights)

        x_k = next_x_k

        # Save data to avoid more communication
        prev_s_is = neighbor_s_is
        prev_gradf_is = neighbor_gradf_is

        if (k+1)/N >= percent:
            if local_gen_id == 1:
                print('[{}%]: '.format(int(percent*100)), flush=True, end='')
            percent += 0.1
            print_final_score(x_k, f_i_idxs, gen_specs, libE_info)

    return None, persis_info, FINISHED_PERSISTENT_GEN_TAG



import os
import logging
from collections import Counter
from collections import OrderedDict
import numpy as np
from libensemble.resources.rset_resources import RSetResources

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


class WorkerResourcesException(Exception):
    """Worker resources exception."""


class ResourceManagerException(Exception):
    """Resource Manager exception."""


class ResourceManager(RSetResources):

    rset_dtype = [('assigned', int),  # Holds worker ID assigned to or zero
                  ('group', int)      # Group ID this resource set belongs to
                  # ('pool', int),    # Pool ID (eg. separate gen/sim resources) - not yet used.
                  ]

    def __init__(self, num_workers, resources):
        """Initializes a new WorkerResources instance

        Determines the compute resources available for current worker, including
        node list and cores/hardware threads available within nodes.

        Parameters
        ----------

        workerID: int
            workerID of current process

        comm: Comm
            The Comm object for manager/worker communications

        resources: Resources
            A Resources object containing global nodelist and intranode information

        """
        super().__init__(num_workers, resources)
        self.index_list = ResourceManager.get_index_list(self.num_workers,
                                                         self.total_num_rsets,
                                                         resources.zero_resource_workers)

        self.rsets = np.zeros(self.total_num_rsets, dtype=ResourceManager.rset_dtype)
        self.rsets['assigned'] = 0
        self.rsets['group'] = ResourceManager.get_group_list(self.split_list)
        self.num_groups = self.rsets['group'][-1]
        self.rsets_free = self.total_num_rsets

        # Useful for scheduling tasks with different sized groups (resource sets per node).
        unique, counts = np.unique(self.rsets['group'], return_counts=True)
        self.group_sizes = dict(zip(unique, counts))
        self.ngroups_by_size = Counter(counts)
        self.even_groups = True if len(self.ngroups_by_size) == 1 else False

    def assign_rsets(self, rset_team, worker_id):
        """Mark the resource sets given by rset_team as assigned to worker_id"""

        if rset_team:
            rteam = self.rsets['assigned'][rset_team]
            for i, wid in enumerate(rteam):
                if wid == 0:
                    self.rsets['assigned'][rset_team[i]] = worker_id
                    self.rsets_free -= 1
                elif wid != worker_id:
                    ResourceManagerException("Error: Attempting to assign rsets {}"
                                             " already assigned to workers: {}".
                                             format(rset_team, rteam))

    def free_rsets(self, worker=None):
        """Free up assigned resource sets"""
        if worker is None:
            self.rsets['assigned'] = 0
            self.rsets_free = self.total_num_rsets
        else:
            rsets_to_free = np.where(self.rsets['assigned'] == worker)[0]
            self.rsets['assigned'][rsets_to_free] = 0
            self.rsets_free += len(rsets_to_free)

    @staticmethod
    def get_group_list(split_list):
        group = 1
        group_list = []
        node = split_list[0]
        for i in range(len(split_list)):
            if split_list[i] == node:
                group_list.append(group)
            else:
                node = split_list[i]
                group += 1
                group_list.append(group)
        return group_list

    @staticmethod
    def get_index_list(num_workers, num_rsets, zero_resource_list):
        """Map WorkerID to index into a nodelist"""
        index = 0
        index_list = []
        for i in range(1, num_workers+1):
            if i in zero_resource_list:
                index_list.append(None)
            else:
                if index >= num_rsets:
                    # Not enough rsets
                    index_list.append(None)
                else:
                    index_list.append(index)
                index += 1
        return index_list


class WorkerResources(RSetResources):
    """Provide system resources per worker to libEnsemble and executor.

    **Object Attributes:**

    Some of these attributes may be updated as the ensemble progresses.

    ``rsets`` below is used to abbreviate ``resource sets``.

    :ivar int workerID: workerID for this worker.
    :ivar list local_nodelist: A list of all nodes assigned to this worker.
    :ivar list rset_team: List of rset IDs currently assigned to this worker.
    :ivar int num_rsets: The number of resource sets assigned to this worker.
    :ivar dict slots: A dictionary with a list of slot IDs for each node.
    :ivar bool even_slots: Determines if the slots evenly divide amongst nodes.
    :ivar int slot_count: The number of slots per node if even_slots is True, else None.
    :ivar list slots_on_node: A list of slots IDs if even_slots is True, else None.
    :ivar int local_node_count: The number of nodes available to this worker (rounded up to whole number).
    :ivar int rsets_per_node: The number of rsets per node (if a rset > 1 node, will be 1).

    The worker_resources attribtues can be queried, and convenience functions
    called, via the resources class attribute. For example:

    With resources imported:

    .. code-block:: python

        from libensemble.resources.resources import Resources

    A user function (sim/gen) may do:

    .. code-block:: python

        resources = Resources.resources.worker_resources
        num_nodes = resources.local_node_count
        cores_per_node = resources.slot_count  # One CPU per GPU
        resources.set_env_to_slots("CUDA_VISIBLE_DEVICES")  # Use convenience function.

    Note that **slots** are resource sets enumerated on a node (starting with zero).
    If a resource set has more than one node, then each node is considered to have slot zero.

    If ``even_slots`` is True, then the attributes ``slot_count`` and ``slots_on_node``
    can be used for simplicity, Otherwise, the ``slots`` dictionary can be used to get
    information for each node.

    """

    def __init__(self, num_workers, resources, workerID):

        """Initializes a new WorkerResources instance

        Determines the compute resources available for current worker, including
        node list and cores/hardware threads available within nodes.

        Parameters
        ----------

        workerID: int
            workerID of current process

        comm: Comm
            The Comm object for manager/worker communications

        resources: Resources
            A Resources object containing global nodelist and intranode information

        """
        super().__init__(num_workers, resources)
        self.workerID = workerID
        self.local_nodelist = []
        self.rset_team = None
        self.num_rsets = 0
        self.slots = None
        self.even_slots = None
        self.slot_count = None
        self.slots_on_node = None
        self.zero_resource_workers = resources.zero_resource_workers
        self.local_node_count = len(self.local_nodelist)
        self.set_slot_count()

    # User convenience functions ----------------------------------------------

    def get_slots_as_string(self, multiplier=1, delimiter=','):
        """Returns list of slots as a string

        :param multiplier: Optional int. Assume this many items per slot.
        :param delimiter: Optional int. Delimiter for output string.
        """

        if self.slots_on_node is None:
            logger.warning("Slots on node is None when requested as a string")
            return None

        n = multiplier
        slot_list = [j for i in self.slots_on_node for j in range(i*n, (i+1)*n)]
        slots = delimiter.join(map(str, slot_list))
        return slots

    def set_env_to_slots(self, env_var, multiplier=1, delimiter=','):
        """Sets the given environment variable to slots

        :param env_var: String. Name of environment variable to set.
        :param multiplier: Optional int. Assume this many items per slot.
        :param delimiter: Optional int. Delimiter for output string.

        Example usage in a sim function:

        With resources imported:

        .. code-block:: python

            from libensemble.resources.resources import Resources

        Obtain worker resources:

        .. code-block:: python

            resources = Resources.resources.worker_resources
            resources.set_env_to_slots("CUDA_VISIBLE_DEVICES")

        """

        os.environ[env_var] = self.get_slots_as_string(multiplier, delimiter)

    # libEnsemble functions ---------------------------------------------------

    def set_rset_team(self, rset_team):
        """Update worker team and local attributes

        Updates: rset_team
                 local_nodelist
                 slots (dictionary with list of partitions for each node)
                 slot_count - number of slots on each node
                 local_node_count
        """
        if self.workerID in self.zero_resource_workers:
            return

        if rset_team != self.rset_team:  # Order matters
            self.rset_team = rset_team
            self.num_rsets = len(rset_team)
            self.local_nodelist, self.slots = \
                WorkerResources.get_local_nodelist(self.workerID, self.rset_team,
                                                   self.split_list, self.rsets_per_node)
            self.set_slot_count()
            self.local_node_count = len(self.local_nodelist)

    def set_slot_count(self):
        if self.slots:
            # Check if same slots on each node (not just lengths)
            first_node_slots = list(self.slots.values())[0]
            all_match = True
            for node_list in self.slots.values():
                if node_list != first_node_slots:
                    all_match = False
                    break

            self.even_slots = True if all_match else False
            if self.even_slots:
                self.slots_on_node = first_node_slots
                self.slot_count = len(self.slots_on_node)
            else:
                self.slots_on_node = None
                self.slot_count = None

    @staticmethod
    def get_local_nodelist(workerID, rset_team, split_list, rsets_per_node):
        """Returns the list of nodes available to the given worker and the slot dictionary"""

        if workerID is None:
            raise WorkerResourcesException("Worker has no workerID - aborting")

        team_list = []
        for index in rset_team:
            team_list += split_list[index]

        local_nodelist = list(OrderedDict.fromkeys(team_list))  # Maintain order of nodes
        logger.debug("Worker's local_nodelist is {}".format(local_nodelist))

        slots = {}
        for node in local_nodelist:
            slots[node] = []

        for index in rset_team:
            mynodes = split_list[index]
            # If rset has > 1 node, then all nodes just have a single slot (slot 0).
            if len(mynodes) > 1:
                for node in mynodes:
                    slots[node].append(0)
            else:
                mynode = split_list[index][0]
                pos_in_node = index % rsets_per_node
                slots[mynode].append(pos_in_node)

        return local_nodelist, slots



"""
Tests libEnsemble's capability to abort after a certain amount of elapsed time.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_6-hump_camel_elapsed_time_abort.py
   python3 test_6-hump_camel_elapsed_time_abort.py --nworkers 3 --comms local
   python3 test_6-hump_camel_elapsed_time_abort.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams, eprint

nworkers, is_manager, libE_specs, _ = parse_args()

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
    'user': {'pause_time': 2},
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, (2,))],
    'user': {
        'gen_batch_size': 5,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {
    'alloc_f': give_sim_work_first,
    'out': [],
    'user': {
        'batch_mode': False,
        'num_active_gens': 2,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'elapsed_wallclock_time': 1}

# Perform the run
H, persis_info, flag = libE(
    sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs, alloc_specs=alloc_specs
)

if is_manager:
    eprint(flag)
    eprint(H)
    assert flag == 2
    save_libE_output(H, persis_info, __file__, nworkers)



"""
Runs libEnsemble with uniform random sampling on the chwirut least squares
problem.  All 214 residual calculations for a given point are performed as a
single simulation evaluation. NaNs are injected probabilistically in order to
test the allocation function's ability to preempt future residual
calculations. Also, the allocation function tries to preempt calculations
corresponding to points with partial sum-squared error worse than the
best-evaluated point so far.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_uniform_sampling_one_residual_at_a_time.py

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi
# TESTSUITE_NPROCS: 2 4

import sys
import numpy as np
from copy import deepcopy

# Import libEnsemble items
from libensemble.libE import libE
from libensemble.sim_funcs.chwirut1 import chwirut_eval as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample_obj_components as gen_f
from libensemble.alloc_funcs.fast_alloc_and_pausing import give_sim_work_first
from libensemble.tests.regression_tests.support import persis_info_3 as persis_info
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()
if libE_specs['comms'] == 'tcp':
    # Can't use the same interface for manager and worker if we want
    # repeated calls to libE -- the manager sets up a different server
    # each time, and the worker will not know what port to connect to.
    sys.exit("Cannot run with tcp when repeated calls to libE -- aborting...")

# Declare the run parameters/functions
m = 214
n = 3
budget = 10 * m

sim_specs = {
    'sim_f': sim_f,
    'in': ['x', 'obj_component'],
    'out': [('f_i', float)],
    'user': {'component_nan_frequency': 0.01},
}

# lb tries to avoid x[1]=-x[2], which results in division by zero in chwirut.
gen_specs = {
    'gen_f': gen_f,
    'in': ['pt_id'],
    'out': [('x', float, n), ('priority', float), ('paused', bool), ('obj_component', int), ('pt_id', int)],
    'user': {
        'gen_batch_size': 2,
        'single_component_at_a_time': True,
        'combine_component_func': lambda x: np.sum(np.power(x, 2)),
        'lb': (-2 - np.pi / 10) * np.ones(n),
        'ub': 2 * np.ones(n),
        'components': m,
    },
}

alloc_specs = {
    'alloc_f': give_sim_work_first,  # Allocation function
    'out': [],  # Output fields (included in History)
    'user': {
        'stop_on_NaNs': True,  # Should alloc preempt evals
        'batch_mode': True,  # Wait until all sim evals are done
        'num_active_gens': 1,  # Only allow one active generator
        'stop_partial_fvec_eval': True,  # Should alloc preempt evals
    },
}
# end_alloc_specs_rst_tag

persis_info = add_unique_random_streams(persis_info, nworkers + 1)
persis_info_safe = deepcopy(persis_info)

exit_criteria = {'sim_max': budget, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)
if is_manager:
    assert flag == 0
    save_libE_output(H, persis_info, __file__, nworkers)

# Perform the run but not stopping on NaNs
alloc_specs['user'].pop('stop_on_NaNs')
persis_info = deepcopy(persis_info_safe)
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)
if is_manager:
    assert flag == 0

# Perform the run also not stopping on partial fvec evals
alloc_specs['user'].pop('stop_partial_fvec_eval')
persis_info = deepcopy(persis_info_safe)
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)
if is_manager:
    assert flag == 0



"""
Tests libEnsemble's inverse_bayes generator function

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_inverse_bayes_example.py
   python3 test_inverse_bayes_example.py --nworkers 3 --comms local
   python3 test_inverse_bayes_example.py --nworkers 3 --comms tcp

Debugging:
   mpiexec -np 4 xterm -e "python3 inverse_bayes_example.py"

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4

import sys
import numpy as np

from libensemble.libE import libE
from libensemble.sim_funcs.inverse_bayes import likelihood_calculator as sim_f
from libensemble.gen_funcs.persistent_inverse_bayes import persistent_updater_after_likelihood as gen_f
from libensemble.alloc_funcs.inverse_bayes_allocf import only_persistent_gens_for_inverse_bayes as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams

# Parse args for test code
nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('like', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': [],
    'out': [('x', float, 2), ('batch', int), ('subbatch', int), ('prior', float), ('prop', float), ('weight', float)],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'subbatch_size': 3,
        'num_subbatches': 2,
        'num_batches': 10,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

# Tell libEnsemble when to stop
val = gen_specs['user']['subbatch_size'] * gen_specs['user']['num_subbatches'] * gen_specs['user']['num_batches']
exit_criteria = {
    'sim_max': val,
    'elapsed_wallclock_time': 300,
}

alloc_specs = {'out': [], 'alloc_f': alloc_f}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert flag == 0
    # Change the last weights to correct values (H is a list on other cores and only array on manager)
    ind = 2 * gen_specs['user']['subbatch_size'] * gen_specs['user']['num_subbatches']
    H[-ind:] = H['prior'][-ind:] + H['like'][-ind:] - H['prop'][-ind:]
    assert len(H) == 60, "Failed"



import numpy as np

from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport


def persistent_updater_after_likelihood(H, persis_info, gen_specs, libE_info):
    """
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)
    subbatch_size = gen_specs['user']['subbatch_size']
    num_subbatches = gen_specs['user']['num_subbatches']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # Receive information from the manager (or a STOP_TAG)
    batch = -1
    tag = None
    w = np.nan
    while tag not in [STOP_TAG, PERSIS_STOP]:
        batch += 1
        H_o = np.zeros(subbatch_size*num_subbatches, dtype=gen_specs['out'])
        if np.all(~np.isnan(w)):
            H_o['weight'] = w
        for j in range(num_subbatches):
            for i in range(subbatch_size):
                row = subbatch_size*j + i
                H_o['x'][row] = persis_info['rand_stream'].uniform(lb, ub, (1, n))
                H_o['subbatch'][row] = j
                H_o['batch'][row] = batch
                H_o['prior'][row] = np.random.randn()
                H_o['prop'][row] = np.random.randn()

        # Send data and get next assignment
        tag, Work, calc_in = ps.send_recv(H_o)
        if calc_in is not None:
            w = H_o['prior'] + calc_in['like'] - H_o['prop']

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG



"""
Runs libEnsemble testing the zero_resource_workers argument with 2 workers per
node.

This test must be run on an odd number of workers >= 3 (e.g. even number of
procs when using mpi4py).

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_zero_resource_workers_subnode.py
   python3 test_zero_resource_workers_subnode.py --nworkers 3 --comms local
   python3 test_zero_resource_workers_subnode.py --nworkers 3 --comms tcp
"""

import sys
import numpy as np

from libensemble.libE import libE
from libensemble.sim_funcs.run_line_check import runline_check as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor
from libensemble.tests.regression_tests.common import create_node_file
from libensemble import logger

# logger.set_level('DEBUG')  # For testing the test
logger.set_level('INFO')

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 4

nworkers, is_manager, libE_specs, _ = parse_args()
rounds = 1
sim_app = '/path/to/fakeapp.x'
comms = libE_specs['comms']

libE_specs['zero_resource_workers'] = [1]
libE_specs['dedicated_mode'] = True
libE_specs['enforce_worker_core_bounds'] = True

# To allow visual checking - log file not used in test
log_file = 'ensemble_zrw_subnode_comms_' + str(comms) + '_wrks_' + str(nworkers) + '.log'
logger.set_filename(log_file)

nodes_per_worker = 0.5

# For varying size test - relate node count to nworkers
in_place = libE_specs['zero_resource_workers']
n_gens = len(in_place)
nsim_workers = nworkers - n_gens

if not (nsim_workers * nodes_per_worker).is_integer():
    sys.exit("Sim workers ({}) must divide evenly into nodes".format(nsim_workers))

comms = libE_specs['comms']
node_file = 'nodelist_zero_resource_workers_subnode_comms_' + str(comms) + '_wrks_' + str(nworkers)
nnodes = int(nsim_workers * nodes_per_worker)

# Mock up system
custom_resources = {
    'cores_on_node': (16, 64),  # Tuple (physical cores, logical cores)
    'node_file': node_file,
}  # Name of file containing a node-list
libE_specs['resource_info'] = custom_resources

if is_manager:
    create_node_file(num_nodes=nnodes, name=node_file)

if comms == 'mpi':
    libE_specs['mpi_comm'].Barrier()

# Create executor and register sim to it.
exctr = MPIExecutor(custom_info={'mpi_runner': 'srun'})
exctr.register_app(full_path=sim_app, calc_type='sim')

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': [],
    'out': [('x', float, (n,))],
    'user': {
        'initial_batch_size': 20,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {'alloc_f': alloc_f, 'out': []}
persis_info = add_unique_random_streams({}, nworkers + 1)
exit_criteria = {'sim_max': (nsim_workers) * rounds}

# Each worker has 2 nodes. Basic test list for portable options
test_list_base = [
    {'testid': 'base1'},  # Give no config and no extra_args
    {'testid': 'base2', 'nprocs': 5},
    {'testid': 'base3', 'nnodes': 1},
    {'testid': 'base4', 'ppn': 6},
]

exp_srun = [
    'srun -w node-1 --ntasks 8 --nodes 1 --ntasks-per-node 8 /path/to/fakeapp.x --testid base1',
    'srun -w node-1 --ntasks 5 --nodes 1 --ntasks-per-node 5 /path/to/fakeapp.x --testid base2',
    'srun -w node-1 --ntasks 8 --nodes 1 --ntasks-per-node 8 /path/to/fakeapp.x --testid base3',
    'srun -w node-1 --ntasks 6 --nodes 1 --ntasks-per-node 6 /path/to/fakeapp.x --testid base4',
]

test_list = test_list_base
exp_list = exp_srun
sim_specs['user'] = {
    'tests': test_list,
    'expect': exp_list,
    'nodes_per_worker': nodes_per_worker,
    'persis_gens': n_gens,
}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

# All asserts are in sim func



"""
libensemble utility class -- manages timer
"""

import datetime


# https://stackoverflow.com/questions/5998245/get-current-time-in-milliseconds-in-python
def TimestampMillisec64():
    return int((datetime.datetime.utcnow() - datetime.datetime(1970, 1, 1)).total_seconds() * 1000)


class Timer:
    """Timer class used in libensemble.

    Attributes
    ----------

    tcum: float:
        Total time recorded by timer.

    tstart: float:
        Most recent starting time.

    tend: float:
        Most recent ending time.

    timing: bool:
        Indicates whether the timer is currently active.
    """

    def __init__(self):
        """Initialize a new timer."""
        self.tcum = 0.0
        self.tstart = 0.0
        self.tend = 0.0
        self.timing = False

    def __str__(self):
        """Return a string representation of the timer."""
        return ("Time: {0:.3f} Start: {1} End: {2}".
                format(self.total, self.date_start, self.date_end))

    @property
    def date_start(self):
        """Return a string representing the start datetime."""
        start_time = datetime.datetime.fromtimestamp(self.tstart / 1000)
        return start_time.strftime("%Y-%m-%d %H:%M:%S") + '.' + str(self.tstart)[-3:]

    @property
    def date_end(self):
        """Return a string representing the end datetime."""
        end_time = datetime.datetime.fromtimestamp(self.tend / 1000)
        return end_time.strftime("%Y-%m-%d %H:%M:%S") + '.' + str(self.tend)[-3:]

    @property
    def elapsed(self):
        """Return time since last start (active) or in most recent interval."""
        etime = self.tend if not self.timing else TimestampMillisec64()
        return (etime - self.tstart) / 1000

    @property
    def total(self):
        """Return the total time since last start."""
        if self.timing:
            return self.tcum / 1000 + self.elapsed  # second term divided above
        return self.tcum / 1000

    def start(self):
        """Start the timer."""
        self.tstart = TimestampMillisec64()
        self.timing = True

    def stop(self):
        """Stop the timer."""
        self.tend = TimestampMillisec64()
        self.timing = False
        self.tcum += (self.tend-self.tstart)

    def __enter__(self):
        """Enter a timing context."""
        self.start()
        return self

    def __exit__(self, etype, value, traceback):
        """Exit a timing context."""
        self.stop()


class TaskTimer(Timer):
    """Timer class used in executor tasks."""
    def __str__(self):
        """Return a string representation of the timer."""
        return ("TaskTime: {0:.2f} JStart: {1} JEnd: {2}".
                format(self.total, self.date_start, self.date_end))



"""
Support for logging over comms
====================================================

The comms mechanism can be used to support consolidated logging at the
manager by an appropriate choice of handlers and filters.  The default
logging behavior is to install a CommLogHandler at each worker, which is
used to pass messages to be handled at the manager, where they are then
selected and emitted.  The WorkerID filter is used to add contextual
information (in the form of a worker field) that identifies the origin of
a given log message (manager or worker ID).
"""

import logging
import sys


class LogConfig:
    """Class for storing logging configuration info"""
    config = None

    def __init__(self, name):
        """Instantiate a new LogConfig instance."""
        LogConfig.config = self
        self.logger_set = False
        self.log_level = logging.INFO
        self.name = name
        self.stats_name = name + ".calc stats"
        self.filename = "ensemble.log"
        self.stat_filename = 'libE_stats.txt'
        self.fmt = '[%(worker)s]  %(asctime)s %(name)s (%(levelname)s): %(message)s'
        self.stderr_level = logging.MANAGER_WARNING

    def set_level(self, level):
        """Set logger level either before or after creating loggers"""
        numeric_level = getattr(logging, level.upper(), 10)
        self.log_level = numeric_level
        if self.logger_set:
            logger = logging.getLogger(self.name)
            logger.setLevel(self.log_level)

    def set_stderr_level(self, level):
        """ Set logger level for copying messages to stderr"""
        numeric_level = getattr(logging, level.upper(), 30)
        self.stderr_level = numeric_level


class CommLogHandler(logging.Handler):
    """Logging handler class that forwards LogRecords to a Comm.
    """

    def __init__(self, comm, pack=None, level=logging.NOTSET):
        """Initialize the handler instance, setting the level and the comm."""
        super().__init__(level)
        self.comm = comm
        self.pack = pack

    def emit(self, record):
        """Actually log the record."""
        if self.pack is not None:
            self.comm.send(*self.pack(record))
        else:
            self.comm.send(record)


class WorkerIDFilter(logging.Filter):
    """Logging filter to add worker ID to records.
    """

    def __init__(self, worker_id):
        super().__init__()
        self.worker_id = worker_id

    def filter(self, record):
        """Add worker ID to a LogRecord"""
        record.worker = getattr(record, 'worker', self.worker_id)
        return True


class ErrorFilter(logging.Filter):
    """ Filter to choose messages for stderr of user-defined level"""

    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record):
        """ Confirm messages that exceed specified level """
        return record.levelno >= self.level


def worker_logging_config(comm, worker_id=None):
    """Add a comm handler with worker ID filter to the indicated logger.
    """

    logconfig = LogConfig.config
    logger = logging.getLogger(logconfig.name)
    ch = CommLogHandler(comm, pack=lambda rec: (0, rec))
    ch.addFilter(WorkerIDFilter(worker_id or comm.rank))

    if logconfig.logger_set:
        for hdl in logger.handlers[:]:
            logger.removeHandler(hdl)
            hdl.close()
    else:
        logger.propagate = False
        logger.setLevel(logconfig.log_level)
        logconfig.logger_set = True

    logger.addHandler(ch)


def manager_logging_config():
    """Add file-based logging at manager.
    """

    # Regular logging
    logconfig = LogConfig.config
    if not logconfig.logger_set:
        formatter = logging.Formatter(logconfig.fmt)
        wfilter = WorkerIDFilter(0)
        fh = logging.FileHandler(logconfig.filename, mode='w')
        fh.addFilter(wfilter)
        fh.setFormatter(formatter)
        logger = logging.getLogger(logconfig.name)
        logger.propagate = False
        logger.setLevel(logconfig.log_level)  # Formatter filters on top of this
        logger.addHandler(fh)
        logconfig.logger_set = True

        # Stats logging
        # NB: Could add a specialized handler for immediate flushing
        fhs = logging.FileHandler(logconfig.stat_filename, mode='w')
        fhs.addFilter(wfilter)
        fhs.setFormatter(logging.Formatter('Worker %(worker)5d: %(message)s'))
        stat_logger = logging.getLogger(logconfig.stats_name)
        stat_logger.propagate = False
        stat_logger.setLevel(logging.DEBUG)
        stat_logger.addHandler(fhs)

        # Mirror error-logging to stderr of user-specified level
        fhe = logging.StreamHandler(stream=sys.stderr)
        fhe.addFilter(wfilter)
        efilter = ErrorFilter(logconfig.stderr_level)
        fhe.addFilter(efilter)
        fhe.setFormatter(formatter)
        logger.addHandler(fhe)

        def close_logs():
            fh.close()
            fhs.close()

        return close_logs



"""
Tests libEnsemble with a simple persistent uniform sampling generator
function.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_uniform_sampling.py
   python3 test_persistent_uniform_sampling.py --nworkers 3 --comms local
   python3 test_persistent_uniform_sampling.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.rosenbrock import rosenbrock_eval as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('grad', float, n)],
}

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['x', 'f', 'grad', 'sim_id'],
    'out': [('x', float, (n,))],
    'user': {
        'initial_batch_size': 20,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)
for i in persis_info:
    persis_info[i]['get_grad'] = True

exit_criteria = {'gen_max': 40, 'elapsed_wallclock_time': 300}

libE_specs['kill_canceled_sims'] = False
# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert len(np.unique(H['gen_time'])) == 2

    save_libE_output(H, persis_info, __file__, nworkers)



"""
Tests CUDA variable resource detection in libEnsemble

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_sampling_CUDA_variable_resources.py

When running with the above command, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi, local
# TESTSUITE_NPROCS: 4

import sys
import numpy as np
import pkg_resources

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_CUDA_variable_resources as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import uniform_random_sample_with_variable_resources as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['zero_resource_workers'] = [1]
libE_specs['sim_dirs_make'] = True
libE_specs['ensemble_dir_path'] = './ensemble_CUDA_variable_w' + str(nworkers)

if libE_specs['comms'] == 'tcp':
    sys.exit("This test only runs with MPI or local -- aborting...")

# Get paths for applications to run
six_hump_camel_app = pkg_resources.resource_filename('libensemble.sim_funcs', 'six_hump_camel.py')
exctr = MPIExecutor()
exctr.register_app(full_path=six_hump_camel_app, app_name='six_hump_camel')

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
    'user': {},
}

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'x', 'sim_id'],
    'out': [('priority', float), ('resource_sets', int), ('x', float, n)],
    'user': {
        'initial_batch_size': nworkers - 1,
        'max_resource_sets': nworkers - 1,  # Any sim created can req. 1 worker up to all.
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {
    'alloc_f': alloc_f,
    'user': {'give_all_with_same_priority': False},
}

persis_info = add_unique_random_streams({}, nworkers + 1)
exit_criteria = {'sim_max': 40, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(
    sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs, alloc_specs=alloc_specs
)

if is_manager:
    assert flag == 0
    save_libE_output(H, persis_info, __file__, nworkers)



#!/usr/bin/env python
import os
import requests
import datetime

from libensemble.libE import libE
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble import logger

# Import our simulator and generator functions
from openmm_md_simf import run_openmm_sim_f
from keras_cvae_ml_genf import run_keras_cvae_ml_genf

# Import DeepDriveMD components for registering with libEnsemble's executor
from deepdrivemd.sim.openmm import run_openmm
from deepdrivemd.aggregation.basic import aggregate
from deepdrivemd.models.keras_cvae import train
from deepdrivemd.selection.latest import select_model
from deepdrivemd.agents.lof import lof

logger.set_level('INFO')


def download_data(url, file):
    if file not in os.listdir('.'):
        print('Downloading ' + file + ' ...')
        out = requests.get(url)
        with open(file, 'wb') as f:
            f.write(out.content)


folded_url = 'https://raw.githubusercontent.com/DeepDriveMD/DeepDriveMD-pipeline/main/data/bba/1FME-folded.pdb'
unfolded_url = folded_url.replace('1FME-folded.pdb', 'system/1FME-unfolded.pdb')

for url in [folded_url, unfolded_url]:
    download_data(url, url.split('/')[-1])

# Parse comms type, number of workers, etc. from command-line
nworkers, is_manager, libE_specs, _ = parse_args()

ddmd_apps = {'molecular_dynamics': run_openmm.__file__,
             'aggregation': aggregate.__file__,
             'machine_learning': train.__file__,
             'model_selection': select_model.__file__,
             'agent': lof.__file__}

# Submit each component via a MPI runner. This Executor can be swapped with the Balsam
#  executor to potentially submit specific components to separate systems.
exctr = MPIExecutor()

for app in ddmd_apps:
    exctr.register_app(full_path=ddmd_apps[app], app_name=app)

# Specify directory structure where user functions will be called
ensemble_directory = os.path.abspath('./ensemble_' + str(datetime.datetime.today()).replace(' ', '_').split('.')[0])

MD_BATCH_SIZE = 12

# Parameterize our simulator function
sim_specs = {'sim_f': run_openmm_sim_f,
             'in': ['stage_id', 'gen_dir_loc', 'initial', 'task_id'],
             'out': [('sim_dir_loc', "<U70"), ('sim_cstat', int)],
             'user': {'sim_kill_minutes': 15,
                      'sim_length_ns': 1.0,
                      'poll_interval': 1,
                      'reference_pdb_file': os.path.abspath(folded_url.split('/')[-1]),
                      'config_file': 'molecular_dynamics.yaml'}
             }

# Parameterize our persistent generator function
gen_specs = {'gen_f': run_keras_cvae_ml_genf,
             'out': [('sim_id', int), ('stage_id', int), ('task_id', int),
                     ('initial', bool), ('gen_dir_loc', "<U70")],
             'user': {'initial_sample_size': MD_BATCH_SIZE,
                      'outliers': MD_BATCH_SIZE,
                      'last_n_h5_files': MD_BATCH_SIZE,  # ML stage
                      'n_most_recent_h5_files': MD_BATCH_SIZE,  # agent stage
                      'n_traj_frames': int(sim_specs['user']['sim_length_ns']*1000),  # agent stage
                      'skip_aggregation': True,
                      'aggregation_kill_minutes': 15,
                      'machine_learning_kill_minutes': 30,
                      'model_selection_kill_minutes': 15,
                      'agent_kill_minutes': 15,
                      }
             }

for app in ddmd_apps:
    gen_specs['user'][app + '_config'] = app + '.yaml'
    gen_specs['out'].append((app + '_cstat', int))

# Parameterize the provided allocation function
alloc_specs = {'alloc_f': alloc_f,
               'user': {'init_sample_size': MD_BATCH_SIZE}}

# Specify when libEnsemble should shut down
exit_criteria = {'sim_max': 120}

# Additional libEnsemble settings to customize our ensemble directory
libE_specs['sim_dirs_make'] = True
libE_specs['sim_input_dir'] = './sim'
libE_specs['sim_dir_symlink_files'] = [os.path.abspath(unfolded_url.split('/')[-1])]

libE_specs['gen_dirs_make'] = True
libE_specs['gen_input_dir'] = './gen'

libE_specs['ensemble_dir_path'] = ensemble_directory

persis_info = add_unique_random_streams({}, nworkers + 1)

# Call libEnsemble
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria,
                            persis_info=persis_info,
                            alloc_specs=alloc_specs,
                            libE_specs=libE_specs)



"""
This file defines the `gen_f` for Bayesian optimization with a Gaussian process.

The `gen_f` is called once by a dedicated worker and only returns at the end
of the whole libEnsemble run.

This `gen_f` is meant to be used with the `alloc_f` function
`only_persistent_gens`
"""

import numpy as np
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport

# import dragonfly Gaussian Process functions
from dragonfly.exd.domains import EuclideanDomain
from dragonfly.exd.experiment_caller import (EuclideanFunctionCaller,
                                             CPFunctionCaller)
from dragonfly.opt.gp_bandit import EuclideanGPBandit, CPGPBandit
from dragonfly.exd.cp_domain_utils import load_config
from argparse import Namespace


def persistent_gp_gen_f(H, persis_info, gen_specs, libE_info):
    """
    Create a Gaussian Process model, update it as new simulation results
    are available, and generate inputs for the next simulations.

    This is a persistent `genf` i.e. this function is called by a dedicated
    worker and does not return until the end of the whole libEnsemble run.
    """
    # Extract bounds of the parameter space, and batch size
    ub_list = gen_specs['user']['ub']
    lb_list = gen_specs['user']['lb']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # Number of points to generate initially
    number_of_gen_points = gen_specs['user']['gen_batch_size']

    # Initialize the dragonfly GP optimizer
    domain = EuclideanDomain([[lo, up] for lo, up in zip(lb_list, ub_list)])
    func_caller = EuclideanFunctionCaller(None, domain)
    opt = EuclideanGPBandit(func_caller, ask_tell_mode=True,
                            options=Namespace(
                                acq='ts', build_new_model_every=number_of_gen_points,
                                init_capital=number_of_gen_points))
    opt.initialise()

    # If there is any past history, feed it to the GP
    if len(H) > 0:
        for i in range(len(H)):
            x = H['x'][i]
            y = H['f'][i]
            opt.tell([(x, -y)])
        # Update hyperparameters
        opt._build_new_model()

    # Receive information from the manager (or a STOP_TAG)
    tag = None
    while tag not in [STOP_TAG, PERSIS_STOP]:

        # Ask the optimizer to generate `batch_size` new points
        # Store this information in the format expected by libE
        H_o = np.zeros(number_of_gen_points, dtype=gen_specs['out'])
        for i in range(number_of_gen_points):
            x = opt.ask()
            H_o['x'][i] = x
            H_o['resource_sets'][i] = 1

        # Send data and get results from finished simulation
        # Blocking call: waits for simulation results to be sent by the manager
        tag, Work, calc_in = ps.send_recv(H_o)
        if calc_in is not None:
            # Check how many simulations have returned
            n = len(calc_in['f'])
            # Update the GP with latest simulation results
            for i in range(n):
                x = calc_in['x'][i]
                y = calc_in['f'][i]
                opt.tell([(x, -y)])
            # Update hyperparameters
            opt._build_new_model()
            # Set the number of points to generate to that number:
            number_of_gen_points = n
        else:
            number_of_gen_points = 0

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG


def persistent_gp_mf_gen_f(H, persis_info, gen_specs, libE_info):
    """
    Create a Gaussian Process model, for multi-fidelity optimization,
    and update it as new simulation results are available, and generate
    inputs for the next simulations.

    This is a persistent `genf` i.e. this function is called by a dedicated
    worker and does not return until the end of the whole libEnsemble run.
    """
    # Extract bounds of the parameter space, and batch size
    ub_list = gen_specs['user']['ub']
    lb_list = gen_specs['user']['lb']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # Fidelity range.
    fidel_range = gen_specs['user']['range']

    # Get fidelity cost function.
    cost_func = gen_specs['user']['cost_func']

    # Number of points to generate initially
    number_of_gen_points = gen_specs['user']['gen_batch_size']

    # Initialize the dragonfly GP optimizer
    domain = EuclideanDomain([[lo, up] for lo, up in zip(lb_list, ub_list)])
    fidel_space = EuclideanDomain([fidel_range])
    func_caller = EuclideanFunctionCaller(None,
                                          raw_domain=domain,
                                          raw_fidel_space=fidel_space,
                                          fidel_cost_func=cost_func,
                                          raw_fidel_to_opt=fidel_range[-1])
    opt = EuclideanGPBandit(func_caller,
                            ask_tell_mode=True,
                            is_mf=True,
                            options=Namespace(acq='ts',
                                              build_new_model_every=number_of_gen_points,
                                              init_capital=number_of_gen_points))
    opt.initialise()

    # If there is any past history, feed it to the GP
    if len(H) > 0:
        for i in range(len(H)):
            x = H['x'][i]
            z = H['z'][i]
            y = H['f'][i]
            opt.tell([([z], x, -y)])
        # Update hyperparameters
        opt._build_new_model()

    # Receive information from the manager (or a STOP_TAG)
    tag = None
    while tag not in [STOP_TAG, PERSIS_STOP]:

        # Ask the optimizer to generate `batch_size` new points
        # Store this information in the format expected by libE
        H_o = np.zeros(number_of_gen_points, dtype=gen_specs['out'])
        for i in range(number_of_gen_points):
            z, input_vector = opt.ask()
            H_o['x'][i] = input_vector
            H_o['z'][i] = z[0]
            H_o['resource_sets'][i] = max(1, int(z[0]/2))

        # Send data and get results from finished simulation
        # Blocking call: waits for simulation results to be sent by the manager
        tag, Work, calc_in = ps.send_recv(H_o)
        if calc_in is not None:
            # Check how many simulations have returned
            n = len(calc_in['f'])
            # Update the GP with latest simulation results
            for i in range(n):
                x = calc_in['x'][i]
                z = calc_in['z'][i]
                y = calc_in['f'][i]
                opt.tell([([z], x, -y)])
            # Update hyperparameters
            opt._build_new_model()
            # Set the number of points to generate to that number:
            number_of_gen_points = n
        else:
            number_of_gen_points = 0

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG


def persistent_gp_mf_disc_gen_f(H, persis_info, gen_specs, libE_info):
    """
    Create a Gaussian Process model, for multi-fidelity optimization
    in a discrete fidelity space, and update it as new simulation results are
    available, and generate inputs for the next simulations.

    This is a persistent `genf` i.e. this function is called by a dedicated
    worker and does not return until the end of the whole libEnsemble run.
    """
    # Extract bounds of the parameter space, and batch size
    ub_list = gen_specs['user']['ub']
    lb_list = gen_specs['user']['lb']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # Multifidelity settings.
    cost_func = gen_specs['user']['cost_func']
    # discrete_fidel = gen_specs['user']['discrete']
    fidel_range = gen_specs['user']['range']

    # Number of points to generate initially.
    number_of_gen_points = gen_specs['user']['gen_batch_size']

    # Create configuration dictionary from which Dragongly will
    # automatically generate the necessary domains and orderings.
    config_params = {}
    config_params['domain'] = []
    for ub, lb in zip(ub_list, lb_list):
        domain_dict = {
            'max': ub,
            'min': lb,
            'type': 'float'
        }
        config_params['domain'].append(domain_dict)
    config_params['fidel_space'] = [{
        'type': 'discrete',
        'items': fidel_range
    }]
    config_params['fidel_to_opt'] = [fidel_range[-1]]
    config = load_config(config_params)

    # Initialize the dragonfly GP optimizer.
    func_caller = CPFunctionCaller(
        None,
        domain=config.domain,
        domain_orderings=config.domain_orderings,
        fidel_space=config.fidel_space,
        fidel_cost_func=cost_func,
        fidel_to_opt=config.fidel_to_opt,
        fidel_space_orderings=config.fidel_space_orderings)
    opt = CPGPBandit(
        func_caller, ask_tell_mode=True, is_mf=True,
        options=Namespace(
            acq='ts', build_new_model_every=number_of_gen_points,
            init_capital=number_of_gen_points))
    opt.initialise()

    # If there is any past history, feed it to the GP
    if len(H) > 0:
        for i in range(len(H)):
            x = H['x'][i]
            z = H['z'][i]
            y = H['f'][i]
            opt.tell([([z], x, -y)])
        # Update hyperparameters
        opt._build_new_model()

    # Receive information from the manager (or a STOP_TAG)
    tag = None
    while tag not in [STOP_TAG, PERSIS_STOP]:

        # Ask the optimizer to generate `batch_size` new points
        # Store this information in the format expected by libE
        H_o = np.zeros(number_of_gen_points, dtype=gen_specs['out'])
        for i in range(number_of_gen_points):
            z, input_vector = opt.ask()
            H_o['x'][i] = input_vector
            H_o['z'][i] = z[0]
            H_o['resource_sets'][i] = max(1, int(z[0]/2))

        # Send data and get results from finished simulation
        # Blocking call: waits for simulation results to be sent by the manager
        tag, Work, calc_in = ps.send_recv(H_o)
        if calc_in is not None:
            # Check how many simulations have returned
            n = len(calc_in['f'])
            # Update the GP with latest simulation results
            for i in range(n):
                x = calc_in['x'][i]
                z = calc_in['z'][i]
                y = calc_in['f'][i]
                opt.tell([([z], x, -y)])
            # Update hyperparameters
            opt._build_new_model()
            # Set the number of points to generate to that number:
            number_of_gen_points = n
        else:
            number_of_gen_points = 0

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG



#!/usr/bin/env python

import sys
import numpy as np
import argparse

desc = "Script to print selected fields of a libEnsemble history array from a file"
example = '''examples:
 ./print_fields.py out1.npy --fields sim_id x f returned

 If no fields are supplied the whole array is printed.

 You can filter by using conditions, but only boolean are supported currently e.g:
 ./print_fields.py out1.npy -f sim_id x f -c given ~returned

 would show lines where given is True and returned is False.
 '''

np.set_printoptions(linewidth=1)

parser = argparse.ArgumentParser(description=desc, epilog=example,
                                 formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument('-f', '--fields', nargs='+', default=[])
parser.add_argument('-c', '--condition', nargs='+', default=[])
parser.add_argument('-s', '--show-fields', dest='show_fields', action='store_true')
parser.add_argument('args', nargs='*', help='*.npy file')
args = parser.parse_args()

fields = args.fields
npfile = args.args
cond = args.condition
show_fields = args.show_fields

if len(npfile) >= 1:
    np_array = np.load(npfile[0])
else:
    parser.print_help()
    sys.exit()

if not fields:
    fields = list(np_array.dtype.names)

if show_fields:
    print('Showing fields:', fields)

if cond:
    fltr = None
    for cn in cond:
        if cn[0] == '~':
            fltr = ~np_array[cn[1:]] if fltr is None else ~np_array[cn[1:]] & fltr
        else:
            fltr = np_array[cn] if fltr is None else np_array[cn] & fltr
    print(np_array[fields][fltr])
else:
    print(np_array[fields])



import numpy as np


def float_x1000(H, persis_info, sim_specs, _):
    """
    Transforms an array and a scalar input and returns values
    """
    output = np.zeros(1, dtype=sim_specs['out'])

    x1 = H['x'][0][0] * 1000.0
    x2 = H['x'][0][1]

    output['arr_vals'].fill(x1)
    output['scal_val'] = x2 + x2/1e7

    return output, persis_info



"""
Tests libEnsemble prox-slide distributed optimization generator function on a
variety of cases, including:
   - Geometric median
   - SVM with l1 regularization

You can specify which problem to test by setting @prob_id in {0,1}.

This call script uses proximal gradient sliding (https://doi.org/10.1007/s10107-015-0955-5)
to solve the following problems. To test, run using, for any p >= 6,
   mpiexec -np p python3 test_persistent_prox_slide.py
   python3 test_persistent_prox_slide.py --nworkers p --comms local

The number gens will be 4.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 6
# TESTSUITE_OS_SKIP: OSX

import sys
import numpy as np
import numpy.linalg as la
import scipy.sparse as spp
import urllib.request

from libensemble.libE import libE
from libensemble.gen_funcs.persistent_prox_slide import opt_slide as gen_f
from libensemble.alloc_funcs.start_persistent_consensus import start_consensus_persistent_gens as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.tools.consensus_subroutines import get_k_reach_chain_matrix, readin_csv, gm_opt, svm_opt

from libensemble.sim_funcs.geomedian import geomedian_eval
from libensemble.sim_funcs.svm import svm_eval

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")
if nworkers < 5:
    sys.exit('This tests requires at least 5 workers (6 MPI processes)...')

num_gens = 4
A = spp.diags([2, 3, 3, 2]) - get_k_reach_chain_matrix(num_gens, 2)
lam_max = np.amax((la.eig(A.todense())[0]).real)

eps = 1e-1

# 0/1: geometric median (0 with local df, 1 with sim), 2: SVM prob_id = 1, 3&4: SVM w/ STOP_TAG
for prob_id in range(0, 4):
    persis_info = {}
    persis_info['A'] = A

    persis_info = add_unique_random_streams(persis_info, nworkers + 1)
    persis_info['gen_params'] = {}

    if prob_id < 3:
        exit_criteria = {'elapsed_wallclock_time': 600}
    else:
        exit_criteria = {'elapsed_wallclock_time': 600, 'sim_max': 1}

    libE_specs['safe_mode'] = False

    if prob_id <= 1:
        persis_info['print_progress'] = 1
        sim_f = geomedian_eval
        m, n = 10, 20
        prob_name = 'Geometric median'
        M = num_gens / (m ** 2)
        N_const = 4
        err_const = 1e2

        np.random.seed(0)
        B = np.array([np.random.normal(loc=10, scale=1.0, size=n) for i in range(m)])
        persis_info['sim_params'] = {'B': B}

        if prob_id == 1:

            def df(x, i):
                b_i = B[i]
                z = x - b_i
                return (1.0 / m) * z / la.norm(z)

            # Setting @f_i_eval and @df_i_eval tells to gen to compute gradients locally
            persis_info['gen_params'] = {'df_i_eval': df}

    if prob_id >= 2:
        if prob_id == 3:
            if is_manager:
                fname = "http://archive.ics.uci.edu/ml/machine-learning-databases/breast-cancer-wisconsin/wdbc.data"
                urllib.request.urlretrieve(fname, "./wdbc.data")

        if libE_specs['comms'] == 'mpi':
            libE_specs['mpi_comm'].Barrier()

        persis_info['print_progress'] = 0
        sim_f = svm_eval
        m, n = 30, 15
        prob_name = 'SVM with l1 regularization'
        if prob_id > 1:
            prob_name += ' w/ stoppage'
        err_const = 1e1
        N_const = 1
        b, X = readin_csv('wdbc.data')
        X = X.T
        c = 0.1

        # reduce problem size
        b = b[:m]
        X = X[:n, :m]
        # Chosen ad-hoc. This is only upper bound on regularizar.
        M = c * ((m) ** 0.5)

        persis_info['sim_params'] = {'X': X, 'b': b, 'c': c, 'reg': 'l1'}

    sim_specs = {
        'sim_f': sim_f,
        'in': ['x', 'obj_component', 'get_grad'],
        'out': [('f_i', float), ('gradf_i', float, (n,))],
    }

    gen_specs = {
        'gen_f': gen_f,
        'out': [
            ('x', float, (n,)),
            ('f_i', float),
            ('eval_pt', bool),  # eval point
            ('consensus_pt', bool),  # does not require a sim
            ('obj_component', int),  # which {f_i} to eval
            ('get_grad', bool),
        ],
        'user': {'lb': -np.zeros(n), 'ub': np.zeros(n)},
    }

    alloc_specs = {
        'alloc_f': alloc_f,
        'user': {'m': m, 'num_gens': num_gens},
    }

    # Include @f_i_eval and @df_i_eval if we want to compute gradient in gen
    persis_info['gen_params'].update(
        {'M': M, 'R': 10 ** 2, 'nu': 1, 'eps': eps, 'D': 2 * n, 'N_const': N_const, 'lam_max': lam_max}
    )

    if is_manager:
        print('=== Optimizing {} ==='.format(prob_name), flush=True)

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

    if is_manager:
        print('=== End algorithm ===', flush=True)

        # check we completed
        assert flag == 0

    if is_manager and prob_id <= 2:
        if prob_id == 0 or prob_id == 1:
            fstar = gm_opt(np.reshape(B, newshape=(-1,)), m)
        elif prob_id == 2:
            fstar = svm_opt(X, b, c, reg='l1')

        # check we have a Laplacian matrix
        assert la.norm(A.dot(np.zeros(A.shape[1]))) < 1e-15, 'Not a Laplacian matrix'

        # compile sum of {f_i} and {x}, and check their values are bounded by O(eps)
        eval_H = H[H['eval_pt']]

        gen_ids = np.unique(eval_H['gen_worker'])
        assert len(gen_ids) == num_gens, 'Gen did not submit any function eval requests'

        x = np.empty(n * num_gens, dtype=float)
        F = 0

        for i, gen_id in enumerate(gen_ids):
            last_eval_idx = np.where(eval_H['gen_worker'] == gen_id)[0][-1]

            f_i = eval_H[last_eval_idx]['f_i']
            x_i = eval_H[last_eval_idx]['x']

            F += f_i
            x[i * n : (i + 1) * n] = x_i

        A_kron_I = spp.kron(A, spp.eye(n))
        consensus_val = np.dot(x, A_kron_I.dot(x))

        assert F - fstar < err_const * eps, 'Error of {:.4e}, expected {:.4e} (assuming f*={:.4e})'.format(
            F - fstar, err_const * eps, fstar
        )
        assert consensus_val < eps, 'Consensus score of {:.4e}, expected {:.4e}\nx={}'.format(consensus_val, eps, x)



# """
# Runs libEnsemble on a function that returns only nan; tests APOSMM functionality
#
# Execute via one of the following commands (e.g. 3 workers):
#    mpiexec -np 4 python3 test_nan_func_aposmm.py
#    python3 test_nan_func_aposmm.py --nworkers 3 --comms local
#    python3 test_nan_func_aposmm.py --nworkers 3 --comms tcp
#
# The number of concurrent evaluations of the objective function will be 4-1=3.
# """

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4
# TESTSUITE_EXTRA: true

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from support import nan_func as sim_f, aposmm_gen_out as gen_out
from libensemble.gen_funcs.old_aposmm import aposmm_logic as gen_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()
n = 2

sim_specs = {'sim_f': sim_f,
             'in': ['x'],
             'out': [('f', float), ('f_i', float)]}

gen_out += [('x', float, n), ('x_on_cube', float, n), ('obj_component', int)]

gen_specs = {'gen_f': gen_f,
             'in': [o[0] for o in gen_out] + ['f', 'f_i', 'returned'],
             'out': gen_out,
             'user': {'initial_sample_size': 5,
                      'lb': -2*np.ones(n),
                      'ub': 2*np.ones(n)}
             }

if nworkers == 3:
    gen_specs['user']['single_component_at_a_time'] = True
    gen_specs['user']['components'] = 1
    gen_specs['user']['combine_component_func'] = np.linalg.norm

persis_info = add_unique_random_streams({}, nworkers + 1)

# Tell libEnsemble when to stop
exit_criteria = {'sim_max': 100, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info,
                            libE_specs=libE_specs)
if is_manager:
    assert flag == 0
    assert np.all(~H['local_pt'])

    save_libE_output(H, persis_info, __file__, nworkers)



#!/usr/bin/env python

import balsam.launcher.dag as dag

dag.BalsamTask.objects.filter(name__contains='outfile').delete()

for job in dag.BalsamTask.objects.filter(name__contains='job_test_balsam'):
    job.update_state('CREATED')
    job.save()



#!/usr/bin/env python

""" Script to set up apps and jobs for balsam tests """

# Note: To see use of command line interface see bash_scripts/setup_balsam_tests.sh script.
#       Currently that script does not create deps between jobs so may run simultaneously
#       This script tests setup within python (could in theory be integrated with job!)

import os
import sys
import subprocess

import balsam.launcher.dag as dag
from balsam.service import models
AppDef = models.ApplicationDefinition


# Ok so more low level - but can interface app stuff in python directly
def add_app(name, exepath, desc):
    """ Add application to database """
    app = AppDef()
    app.name = name
    app.executable = exepath    # “/full/path/to/python/interpreter /full/path/to/script.py"
    app.description = desc
    # app.default_preprocess = '' # optional
    # app.default_postprocess = '' # optional
    app.save()


# As balsam req python 3.6 lets use subprocess.run
# For any stuff requiring CLI
def run_cmd(cmd, echo=False):
    """ Run a bash command """
    if echo:
        print("\nRunning %s ...\n" % cmd)
    try:
        subprocess.run(cmd.split(), check=True)
    except Exception as e:
        print(e)
        raise("Error: Command %s failed to run" % cmd)


# Use relative paths to balsam_tests dir
work_dir = os.path.dirname(os.path.abspath(__file__))

sim_input_dir = os.path.abspath('../../examples/sim_funcs')
sim_app = "helloworld.py"

# For host code
num_nodes = 1
procs_per_node = 4

job_list = ['test_balsam_1__runjobs.py',
            'test_balsam_2__workerkill.py',
            'test_balsam_3__managerkill.py']


# Currently think only CLI interface for this stuff??
# Check for empty database if poss
run_cmd("balsam rm apps --all", True)
run_cmd("balsam rm jobs --all", True)

# Add user apps - eg helloworld.py
sim_app_name = os.path.splitext(sim_app)[0]  # rm .py extension
sim_app_path = os.path.join(sim_input_dir, sim_app)  # Full path
sim_app_desc = 'Run ' + sim_app_name
run_line = sys.executable + ' ' + sim_app_path
add_app(sim_app_name, run_line, sim_app_desc)


# Add test jobs apps and jobs - and set to run one at a time
prev_job_name = None

for job in job_list:

    app_name = os.path.splitext(job)[0]
    app_path = os.path.join(work_dir, job)
    app_desc = 'Run ' + app_name
    run_line = sys.executable + ' ' + app_path
    add_app(app_name, run_line, app_desc)

    job_name = 'job_' + app_name
    dag.add_job(name=job_name,
                workflow="libe_workflow",
                application=app_name,
                num_nodes=num_nodes,
                procs_per_node=procs_per_node,
                stage_out_url="local:" + work_dir,
                stage_out_files=job_name + ".out")

    # Add dependency between jobs so run one at a time.
    if prev_job_name:
        BalsamJob = dag.BalsamJob
        parent = BalsamJob.objects.get(name=prev_job_name)
        child = BalsamJob.objects.get(name=job_name)
        dag.add_dependency(parent, child)

    prev_job_name = job_name


# Check how to do in API - until then use CLI
run_cmd("balsam ls apps", True)
run_cmd("balsam ls jobs", True)

print("")
run_cmd("echo -e To launch jobs run: balsam launcher --consume-all")
print("")



"""
MPI-based bidirectional communicator
------------------------------------

"""

import time
from mpi4py import MPI
from libensemble.comms.comms import Comm, Timeout


class MPIComm(Comm):
    """MPI-based bidirectional communicator.

    The MPIComm object provides the same message queue pair abstraction as the
    other comm objects.  In order to provide nonblocking queue semantics
    (i.e., send returns immediately after putting the message in an internal
    queue rather than after it is 'in the system'), we use the MPI isend.
    Because of how mpi4py is implemented, the isend call immediately creates
    a pickle in a separate buffer, which is associated with the request
    object, and so will not be garbage-collected until the request object
    goes away.  We keep an internal _outbox list to track all pending
    request objects; if the MPIComm is ever garbage-collected while requests
    are still pending, we cancel those requests.
    """

    def __init__(self, mpi_comm, remote_rank=0):
        """Initialize with a given MPI communicator and rank for the other end"""
        self.mpi_comm = mpi_comm
        self.remote_rank = remote_rank
        self.status = MPI.Status()
        self._outbox = []
        self.recv_buffer = None

    def __del__(self):
        """Wait on anything pending if comm is killed."""
        for req in self._outbox:
            req.Wait()

    def mail_flag(self):
        if self.recv_buffer is not None:
            return True
        # Loop a few times to ensure MPI progress
        for i in range(4):
            if self.mpi_comm.Iprobe(source=self.remote_rank):
                return True
        return False

    def kill_pending(self):
        """Make sure pending requests are cancelled if the comm is killed."""
        for req in self._outbox:
            if not req.Test():
                req.Cancel()
        self._outbox = []

    @property
    def rank(self):
        return self.mpi_comm.Get_rank()

    def clean_outbox(self):
        """Discard the request objects for any completed isends"""
        self._outbox = [req for req in self._outbox if not req.Test()]

    def send(self, *args):
        """Send the requested message (as a pickle) via an MPI isend"""
        self.clean_outbox()
        msg, tag = self.process_outgoing(args)
        req = self.mpi_comm.isend(msg, dest=self.remote_rank, tag=tag)
        self._outbox.append(req)

    def recv(self, timeout=None):
        """Receive a message or raise TimeoutError."""
        if self.recv_buffer is not None:
            result = self.recv_buffer
            self.recv_buffer = None
            return result
        if timeout is not None:
            tfinal = time.time() + timeout
            while not self.mail_flag():
                if time.time() > tfinal:
                    raise Timeout()
        result = self.mpi_comm.recv(source=self.remote_rank, status=self.status)
        return self.process_incoming(result, self.status)

    def process_outgoing(self, msg):
        """Convert a communicator-format message to an MPI message and tag."""
        return msg, 0

    def process_incoming(self, msg, status):
        """Convert an MPI message and tag to a local communicator format message."""
        return msg[0]

    def push_to_buffer(self, *args):
        assert self.recv_buffer is None, "Cannot push back multiple messages"
        self.recv_buffer = args

    def get_num_workers(self):
        return self.mpi_comm.Get_size() - 1


class MainMPIComm(MPIComm):
    """MPI communicator used by the workers and managers for the moment."""

    def process_incoming(self, msg, status):
        return status.Get_tag(), msg

    def process_outgoing(self, msg):
        return msg[1], msg[0]



from libensemble.executors.executor import Executor
from libensemble.executors.mpi_executor import MPIExecutor

import os
import sys
if 'BALSAM_DB_PATH' in os.environ and int(sys.version[2]) >= 6:
    from libensemble.executors.balsam_executor import BalsamMPIExecutor

__all__ = ['BalsamMPIExecutor', 'Executor', 'MPIExecutor']



import numpy as np
from scipy.spatial.distance import cdist


def test_cdist_issue():
    """There is an issue (at least in scipy 1.1.0) with cdist segfaulting."""

    H = np.zeros(20, dtype=[('x', '<f8', (2,)), ('m', '<i8'), ('a', '<f8'), ('b', '?'), ('c', '?'),
                            ('d', '<f8'), ('e', '<f8'), ('fa', '<f8'), ('g', '<i8'), ('h', '<i8'), ('i', '?'),
                            ('j', '<i8'), ('k', '?'), ('f', '<f8'), ('l', '?')])
    np.random.seed(1)
    H['x'] = np.random.uniform(0, 1, (20, 2))
    dist_1 = cdist(np.atleast_2d(H['x'][3]), H['x'], 'euclidean')
    assert len(dist_1), "We didn't segfault"


def test_save():
    """Seeing if I can save parts of the H array."""

    from libensemble.tests.regression_tests.support import uniform_or_localopt_gen_out as gen_out

    n = 2
    gen_out += [('x', float, n), ('x_on_cube', float, n)]
    H = np.zeros(20, dtype=gen_out + [('f', float), ('grad', float, n)])
    np.random.seed(1)
    H['x'] = np.random.uniform(0, 1, (20, 2))
    np.save('H_test', H[['x', 'f', 'grad']])

    assert 1, "We saved correctly"


if __name__ == "__main__":
    test_cdist_issue()
    test_save()



import numpy as np
import logging
from libensemble.message_numbers import EVAL_SIM_TAG, EVAL_GEN_TAG
from libensemble.resources.resources import Resources
from libensemble.resources.scheduler import ResourceScheduler, InsufficientFreeResources  # noqa: F401
from libensemble.output_directory import EnsembleDirectory

logger = logging.getLogger(__name__)
# For debug messages - uncomment
# logger.setLevel(logging.DEBUG)


class AllocException(Exception):
    """Raised for any exception in the alloc support"""


class AllocSupport:
    """A helper class to assist with writing allocation functions.

    This class contains methods for common operations like populating work
    units, determining which workers are available, evaluating what values
    need to be distributed to workers, and others.

    Note that since the ``alloc_f`` is called periodically by the Manager, this
    class instance (if used) will be recreated/destroyed on each loop.
    """

    gen_counter = 0

    def __init__(self, W, manage_resources=False, persis_info={}, scheduler_opts={},
                 user_resources=None, user_scheduler=None):
        """Instantiate a new AllocSupport instance

        ``W`` is. They are referenced by the various methods,
        but are never modified.

        By default, an ``AllocSupport`` instance uses any initiated libEnsemble resource
        module and the built-in libEnsemble scheduler.

        :param W: A :doc:`Worker array<../data_structures/worker_array>`
        :param manage_resources: Optional, Boolean for if to assign resource sets when creating work units
        :param persis_info: Optional, A :doc:`dictionary of persistent information.<../data_structures/libE_specs>`
        :param scheduler_opts: Optional, A dictionary of options to pass to the resource scheduler.
        :param user_resources: Optional, A user supplied ``resources`` object.
        :param user_scheduler: Optional, A user supplied ``user_scheduler`` object.
        """
        self.W = W
        self.persis_info = persis_info
        self.manage_resources = manage_resources
        self.resources = user_resources or Resources.resources
        self.sched = None
        if self.resources is not None:
            wrk_resources = self.resources.resource_manager
            self.sched = user_scheduler or ResourceScheduler(wrk_resources, scheduler_opts)

    def assign_resources(self, rsets_req):
        """Schedule resource sets to a work record if possible.

        For default scheduler, if more than one group (node) is required,
        will try to find even split, otherwise allocates whole nodes.

        Raises ``InsufficientFreeResources`` if the required resources are not
        currently available, or ``InsufficientResourcesError`` if the required
        resources do not exist.

        :param rsets_req: Int. Number of resource sets to request.
        :returns: List of Integers. Resource set indices assigned.
        """
        rset_team = None
        if self.resources is not None:
            rset_team = self.sched.assign_resources(rsets_req)
        return rset_team

    def avail_worker_ids(self, persistent=None, active_recv=False, zero_resource_workers=None):
        """Returns available workers as a list of IDs, filtered by the given options.

        :param persistent: Optional int. Only return workers with given ``persis_state`` (1=sim, 2=gen).
        :param active_recv: Optional Boolean. Only return workers with given active_recv state.
        :param zero_resource_workers: Optional Boolean. Only return workers that require no resources
        :returns: List of worker IDs

        If there are no zero resource workers defined, then the ``zero_resource_workers`` argument will
        be ignored.
        """
        def fltr(wrk, field, option):
            """Filter by condition if supplied"""
            if option is None:
                return True
            return wrk[field] == option

        # For abbrev.
        def fltr_persis():
            if persistent is None:
                return True
            return wrk['persis_state'] == persistent

        def fltr_zrw():
            # If none exist or you did not ask for zrw then return True
            if no_zrw or zero_resource_workers is None:
                return True
            return wrk['zero_resource_worker'] == zero_resource_workers

        def fltr_recving():
            if active_recv:
                return wrk['active_recv']
            else:
                return not wrk['active']

        if active_recv and not persistent:
            raise AllocException("Cannot ask for non-persistent active receive workers")

        # If there are no zero resource workers - then ignore zrw (i.e. use only if they exist)
        no_zrw = not any(self.W['zero_resource_worker'])
        wrks = []
        for wrk in self.W:
            if fltr_recving() and fltr_persis() and fltr_zrw():
                wrks.append(wrk['worker_id'])
        return wrks

    def count_gens(self):
        """Returns the number of active generators."""
        return sum(self.W['active'] == EVAL_GEN_TAG)

    def test_any_gen(self):
        """Returns ``True`` if a generator worker is active."""
        return any(self.W['active'] == EVAL_GEN_TAG)

    def count_persis_gens(self):
        """Return the number of active persistent generators."""
        return sum(self.W['persis_state'] == EVAL_GEN_TAG)

    def _update_rset_team(self, libE_info, wid, H=None, H_rows=None):
        if self.manage_resources and not libE_info.get('rset_team'):
            if self.W[wid-1]['persis_state']:
                return []  # Even if empty list, non-None rset_team stops manager giving default resources
            else:
                if H is not None and H_rows is not None:
                    if 'resource_sets' in H.dtype.names:
                        num_rsets_req = (np.max(H[H_rows]['resource_sets']))  # sim rsets
                    else:
                        num_rsets_req = 1
                else:
                    num_rsets_req = self.persis_info.get('gen_resources', 0)
                return self.assign_resources(num_rsets_req)

    def sim_work(self, wid, H, H_fields, H_rows, persis_info, **libE_info):
        """Add sim work record to given ``Work`` dictionary.

         Includes evaluation of required resources if the worker is not in a
         persistent state.

        :param wid: Int. Worker ID.
        :param H: :doc:`History array<../data_structures/history_array>`. For parsing out requested resource sets.
        :param H_fields: Which fields from :ref:`H<datastruct-history-array>` to send
        :param H_rows: Which rows of ``H`` to send.
        :param persis_info: Worker specific :ref:`persis_info<datastruct-persis-info>` dictionary

        :returns: a Work entry

        Additional passed parameters are inserted into ``libE_info`` in the resulting work record.

        If ``rset_team`` is passed as an additional parameter, it will be honored, assuming that
        any resource checking has already been done.

        """
        libE_info['rset_team'] = self._update_rset_team(libE_info, wid,
                                                        H=H, H_rows=H_rows)  # to parse out resource_sets
        H_fields = AllocSupport._check_H_fields(H_fields)
        libE_info['H_rows'] = AllocSupport._check_H_rows(H_rows)

        work = {'H_fields': H_fields,
                'persis_info': persis_info,
                'tag': EVAL_SIM_TAG,
                'libE_info': libE_info}

        logger.debug("Alloc func packing SIM work for worker {}. Packing sim_ids: {}".
                     format(wid, EnsembleDirectory.extract_H_ranges(work) or None))
        return work

    def gen_work(self, wid, H_fields, H_rows, persis_info, **libE_info):
        """Add gen work record to given ``Work`` dictionary.

         Includes evaluation of required resources if the worker is not in a
         persistent state.

        :param Work: :doc:`Work dictionary<../data_structures/work_dict>`
        :param wid: Worker ID.
        :param H_fields: Which fields from :ref:`H<datastruct-history-array>` to send
        :param H_rows: Which rows of ``H`` to send.
        :param persis_info: Worker specific :ref:`persis_info<datastruct-persis-info>` dictionary

        :returns: A Work entry

        Additional passed parameters are inserted into ``libE_info`` in the resulting work record.

        If ``rset_team`` is passed as an additional parameter, it will be honored, and assume that
        any resource checking has already been done. For example, passing ``rset_team=[]``, would
        ensure that no resources are assigned.
        """

        libE_info['rset_team'] = self._update_rset_team(libE_info, wid)

        if not self.W[wid-1]['persis_state']:
            AllocSupport.gen_counter += 1  # Count total gens
            libE_info['gen_count'] = AllocSupport.gen_counter

        H_fields = AllocSupport._check_H_fields(H_fields)
        libE_info['H_rows'] = AllocSupport._check_H_rows(H_rows)

        work = {'H_fields': H_fields,
                'persis_info': persis_info,
                'tag': EVAL_GEN_TAG,
                'libE_info': libE_info}

        logger.debug("Alloc func packing GEN work for worker {}. Packing sim_ids: {}".
                     format(wid, EnsembleDirectory.extract_H_ranges(work) or None))
        return work

    def _filter_points(self, H_in, pt_filter, low_bound):
        """Returns H and pt_filter filted by lower bound

        :param pt_filter: Optional boolean array filtering expected returned points in ``H``.
        :param low_bound: Optional lower bound for testing all returned.
        """
        # Faster not to slice when whole array
        if low_bound is not None:
            H = H_in[low_bound:]
        else:
            H = H_in

        if pt_filter is None:
            pfilter = True  # Scalar expansion
        else:
            if low_bound is not None:
                pfilter = pt_filter[low_bound:]
            else:
                pfilter = pt_filter
        return H, pfilter

    def all_given(self, H, pt_filter=None, low_bound=None):
        """Returns ``True`` if all expected points have been given to sim

        Excludes cancelled points.

        :param pt_filter: Optional boolean array filtering expected returned points in ``H``.
        :param low_bound: Optional lower bound for testing all returned.
        :returns: True if all expected points have been returned
        """
        H, pfilter = self._filter_points(H, pt_filter, low_bound)
        excluded_points = H['cancel_requested']
        return np.all(H['given'][pfilter & ~excluded_points])

    def all_returned(self, H, pt_filter=None, low_bound=None):
        """Returns ``True`` if all expected points have returned from sim

        Excludes cancelled points that were not already given out.

        :param pt_filter: Optional boolean array filtering expected returned points in ``H``.
        :param low_bound: Optional lower bound for testing all returned.
        :returns: True if all expected points have been returned
        """
        H, pfilter = self._filter_points(H, pt_filter, low_bound)
        excluded_points = H['cancel_requested'] & ~H['given']
        return np.all(H['returned'][pfilter & ~excluded_points])

    def all_given_back(self, H, pt_filter=None, low_bound=None):
        """Returns ``True`` if all expected points have been given back to gen.

        Excludes cancelled points that were not already given out.

        :param pt_filter: Optional boolean array filtering expected returned points in ``H``.
        :param low_bound: Optional lower bound for testing all returned.
        :returns: True if all expected points have been returned
        """
        H, pfilter = self._filter_points(H, pt_filter, low_bound)
        excluded_points = H['cancel_requested'] & ~H['given']
        return np.all(H['given_back'][pfilter & ~excluded_points])

    def points_by_priority(self, H, points_avail, batch=False):
        """Returns indices of points to give by priority

        :param points_avail: Indices of points that are available to give
        :param batch: Optional Boolean. Should batches of points with the same priority be given simultaneously.
        :returns: An array of point indices to give.
        """
        if 'priority' in H.dtype.fields:
            priorities = H['priority'][points_avail]
            if batch:
                q_inds = (priorities == np.max(priorities))
            else:
                q_inds = np.argmax(priorities)
        else:
            q_inds = 0
        return np.nonzero(points_avail)[0][q_inds]

    @staticmethod
    def _check_H_rows(H_rows):
        """Ensure H_rows is a numpy array.  If it is not, then convert if possible,
        else raise an error.

        :returns: ndarray. H_rows
        """
        H_rows = np.atleast_1d(H_rows)  # Makes sure a numpy scalar is an ndarray

        if isinstance(H_rows, np.ndarray):
            return H_rows
        try:
            H_rows = np.fromiter(H_rows, int)
        except Exception:
            raise AllocException("H_rows could not be converted to a numpy array. Type {}".
                                 format(type(H_rows)))
        return H_rows

    @staticmethod
    def _check_H_fields(H_fields):
        """Ensure no duplicates in H_fields"""
        if len(H_fields) != len(set(H_fields)):
            logger.debug("Removing duplicate field(s) when packing work request. {}".format(H_fields))
            H_fields = list(set(H_fields))
            # H_fields = list(OrderedDict.fromkeys(H_fields))  # Maintain order
        return H_fields



#!/usr/bin/env python
"""
Parallel Hello World
"""

from mpi4py import MPI
import sys

size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()
name = MPI.Get_processor_name()

sys.stdout.write("Hello, World! I am process %d of %d on %s.\n" % (rank, size, name))



import numpy as np


def EvaluateFunction(theta, component, X, b, c, reg):
    """
    Evaluates svm with l1 regularization
    """
    i = component
    m = len(b)

    b_i = b[i]
    X_i = X[:, i]

    f_i = max(0, 1-b_i*np.dot(X_i, theta))

    assert reg == 'l1', "Only l1 regularization allowed"

    # if reg is None:
    #     reg_val = 0
    # elif reg == 'l1':
    reg_val = (c/m) * np.sum(np.abs(theta))
    # else:
    #     reg_val = (c/m) * np.dot(theta, theta)

    return f_i+reg_val


def EvaluateJacobian(theta, component, X, b, c, reg):
    """
    Evaluates svm with l1 regularization
    """

    i = component
    m = len(b)

    b_i = b[i]
    X_i = X[:, i]

    score = b_i*np.dot(X_i, theta)

    if score >= 1:
        df_i = np.zeros(len(theta))
    else:
        df_i = -b_i * X_i

    assert reg == 'l1', "Only l1 regularization allowed"

    # if reg is None:
    #     reg_val = 0
    # elif reg == 'l1':
    reg_val = (c/m) * np.sign(theta)
    # else:
    #     reg_val = (2*c/m) * theta

    return df_i+reg_val


def svm_eval(H, persis_info, sim_specs, _):

    X = persis_info['params']['X']
    b = persis_info['params']['b']
    c = persis_info['params']['c']
    reg = persis_info['params'].get('reg', None)

    assert (reg is None) or (reg == 'l1') or (reg == 'l2'), \
        'Incompatable regularization {}'.format(reg)

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        obj_component = H['obj_component'][i]  # which f_i

        if H[i]['get_grad']:
            H_o['gradf_i'][i] = EvaluateJacobian(x, obj_component, X, b, c, reg)
        else:
            H_o['f_i'][i] = EvaluateFunction(x, obj_component, X, b, c, reg)

    return H_o, persis_info



"""
Tests libEnsemble's capability to take in an existing sample of points with
sim_f values and do additional evaluations.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_evaluate_mixed_sample.py
   python3 test_evaluate_mixed_sample.py --nworkers 3 --comms local
   python3 test_evaluate_mixed_sample.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.borehole import borehole as sim_f, gen_borehole_input, borehole_func
from libensemble.alloc_funcs.give_pregenerated_work import give_pregenerated_sim_work as alloc_f
from libensemble.tools import parse_args, save_libE_output

nworkers, is_manager, libE_specs, _ = parse_args()

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float, 8)],
}

gen_specs = {}

n_samp = 1000
n = 8

H0 = np.zeros(n_samp, dtype=[('x', float, 8), ('f', float, 8), ('sim_id', int), ('given', bool), ('returned', bool)])

np.random.seed(0)
H0['x'] = gen_borehole_input(n_samp)

for i in range(500):
    H0['f'][i] = borehole_func(H0['x'][i])

H0['given'][:500] = True
H0['returned'][:500] = True

alloc_specs = {'alloc_f': alloc_f, 'out': [('x', float, n)]}

exit_criteria = {'sim_max': len(H0)}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, alloc_specs=alloc_specs, libE_specs=libE_specs, H0=H0)

if is_manager:
    assert len(H) == len(H0)
    assert np.array_equal(H0['x'], H['x'])
    assert np.all(H['returned'])
    print("\nlibEnsemble correctly didn't add anything to initial sample")
    save_libE_output(H, persis_info, __file__, nworkers)



from libensemble.resources import mpi_resources
from libensemble.executors.executor import jassert
import argparse
import logging

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


class MPIRunner:

    @staticmethod
    def get_runner(mpi_runner_type, runner_name=None):

        mpi_runners = {
            'mpich': MPICH_MPIRunner,
            'openmpi': OPENMPI_MPIRunner,
            'aprun': APRUN_MPIRunner,
            'srun': SRUN_MPIRunner,
            'jsrun': JSRUN_MPIRunner,
            'custom': MPIRunner
        }
        mpi_runner = mpi_runners[mpi_runner_type]
        if runner_name is not None:
            runner = mpi_runner(runner_name)
        else:
            runner = mpi_runner()
        return runner

    def __init__(self, run_command='mpiexec'):
        self.run_command = run_command
        self.mpi_command = [self.run_command, '{extra_args}']
        self.subgroup_launch = False
        self.mfile_support = False
        self.arg_nprocs = ('--LIBE_NPROCS_ARG_EMPTY',)
        self.arg_nnodes = ('--LIBE_NNODES_ARG_EMPTY',)
        self.arg_ppn = ('--LIBE_PPN_ARG_EMPTY',)

    def _get_parser(self, p_args, nprocs, nnodes, ppn):
        parser = argparse.ArgumentParser(description='Parse extra_args', allow_abbrev=False)
        parser.add_argument(*nprocs, type=int, dest='num_procs', default=None)
        parser.add_argument(*nnodes, type=int, dest='num_nodes', default=None)
        parser.add_argument(*ppn, type=int, dest='procs_per_node', default=None)
        args, _ = parser.parse_known_args(p_args)
        return args

    def _parse_extra_args(self, num_procs, num_nodes, procs_per_node,
                          hyperthreads, extra_args):

        splt_extra_args = extra_args.split()
        p_args = self._get_parser(splt_extra_args, self.arg_nprocs, self.arg_nnodes, self.arg_ppn)

        # Only fill from extra_args if not set by portable options
        if num_procs is None:
            num_procs = p_args.num_procs
        if num_nodes is None:
            num_nodes = p_args.num_nodes
        if procs_per_node is None:
            procs_per_node = p_args.procs_per_node

        extra_args = ' '.join(splt_extra_args)
        return num_procs, num_nodes, procs_per_node, p_args

    def _rm_replicated_args(self, num_procs, num_nodes, procs_per_node, p_args):
        if p_args.num_procs is not None:
            num_procs = None
        if p_args.num_nodes is not None:
            num_nodes = None
        if p_args.procs_per_node is not None:
            procs_per_node = None
        return num_procs, num_nodes, procs_per_node

    def express_spec(self, task, num_procs, num_nodes,
                     procs_per_node, machinefile,
                     hyperthreads, extra_args,
                     resources, workerID):

        hostlist = None
        machinefile = None
        # Always use host lists (unless uneven mapping)
        hostlist = mpi_resources.get_hostlist(resources, num_nodes)
        return hostlist, machinefile

    def get_mpi_specs(self, task, num_procs, num_nodes,
                      procs_per_node, machinefile,
                      hyperthreads, extra_args,
                      resources, workerID):
        "Form the mpi_specs dictionary."

        # Return auto_resource variables inc. extra_args additions
        if extra_args:
            num_procs, num_nodes, procs_per_node, p_args = \
                self._parse_extra_args(num_procs, num_nodes, procs_per_node,
                                       hyperthreads, extra_args=extra_args)

        hostlist = None
        if machinefile and not self.mfile_support:
            logger.warning('User machinefile ignored - not supported by {}'.format(self.run_command))
            machinefile = None

        if machinefile is None and resources is not None:
            num_procs, num_nodes, procs_per_node = \
                mpi_resources.get_resources(resources, num_procs, num_nodes,
                                            procs_per_node, hyperthreads)
            hostlist, machinefile = \
                self.express_spec(task, num_procs, num_nodes,
                                  procs_per_node, machinefile,
                                  hyperthreads, extra_args,
                                  resources, workerID)
        else:
            num_procs, num_nodes, procs_per_node = \
                mpi_resources.task_partition(num_procs, num_nodes,
                                             procs_per_node, machinefile)

        # Remove portable variable if in extra_args
        if extra_args:
            num_procs, num_nodes, procs_per_node = \
                self._rm_replicated_args(num_procs, num_nodes,
                                         procs_per_node, p_args)

        return {'num_procs': num_procs,
                'num_nodes': num_nodes,
                'procs_per_node': procs_per_node,
                'extra_args': extra_args,
                'machinefile': machinefile,
                'hostlist': hostlist}


class MPICH_MPIRunner(MPIRunner):

    def __init__(self, run_command='mpirun'):
        self.run_command = run_command
        self.subgroup_launch = True
        self.mfile_support = True
        self.arg_nprocs = ('-n', '-np')
        self.arg_nnodes = ('--LIBE_NNODES_ARG_EMPTY',)
        self.arg_ppn = ('--ppn',)
        self.mpi_command = [self.run_command, '--env {env}',
                            '-machinefile {machinefile}',
                            '-hosts {hostlist}', '-np {num_procs}',
                            '--ppn {procs_per_node}', '{extra_args}']


class OPENMPI_MPIRunner(MPIRunner):

    def __init__(self, run_command='mpirun'):
        self.run_command = run_command
        self.subgroup_launch = True
        self.mfile_support = True
        self.arg_nprocs = ('-n', '-np', '-c', '--n')
        self.arg_nnodes = ('--LIBE_NNODES_ARG_EMPTY',)
        self.arg_ppn = ('-npernode',)
        self.mpi_command = [self.run_command, '-x {env}',
                            '-machinefile {machinefile}',
                            '-host {hostlist}', '-np {num_procs}',
                            '-npernode {procs_per_node}', '{extra_args}']

    def express_spec(self, task, num_procs, num_nodes,
                     procs_per_node, machinefile,
                     hyperthreads, extra_args,
                     resources, workerID):

        hostlist = None
        machinefile = None
        # Use machine files for OpenMPI
        # as "-host" requires entry for every rank

        machinefile = "machinefile_autogen"
        if workerID is not None:
            machinefile += "_for_worker_{}".format(workerID)
        machinefile += "_task_{}".format(task.id)
        mfile_created, num_procs, num_nodes, procs_per_node = \
            mpi_resources.create_machinefile(resources, machinefile,
                                             num_procs, num_nodes,
                                             procs_per_node, hyperthreads)
        jassert(mfile_created, "Auto-creation of machinefile failed")

        return hostlist, machinefile


class APRUN_MPIRunner(MPIRunner):

    def __init__(self, run_command='aprun'):
        self.run_command = run_command
        self.subgroup_launch = False
        self.mfile_support = False
        self.arg_nprocs = ('-n',)
        self.arg_nnodes = ('--LIBE_NNODES_ARG_EMPTY',)
        self.arg_ppn = ('-N',)
        self.mpi_command = [self.run_command, '-e {env}',
                            '-L {hostlist}', '-n {num_procs}',
                            '-N {procs_per_node}', '{extra_args}']


class SRUN_MPIRunner(MPIRunner):

    def __init__(self, run_command='srun'):
        self.run_command = run_command
        self.subgroup_launch = False
        self.mfile_support = False
        self.arg_nprocs = ('-n', '--ntasks')
        self.arg_nnodes = ('-N', '--nodes')
        self.arg_ppn = ('--ntasks-per-node',)
        self.mpi_command = [self.run_command, '-w {hostlist}',
                            '--ntasks {num_procs}',
                            '--nodes {num_nodes}',
                            '--ntasks-per-node {procs_per_node}',
                            '{extra_args}']


class JSRUN_MPIRunner(MPIRunner):

    def __init__(self, run_command='jsrun'):
        self.run_command = run_command
        self.subgroup_launch = True
        self.mfile_support = False

        # TODO: Add multiplier to resources checks (for -c/-a)
        self.arg_nprocs = ('--np', '-n')
        self.arg_nnodes = ('--LIBE_NNODES_ARG_EMPTY',)
        self.arg_ppn = ('-r',)
        self.mpi_command = [self.run_command, '-n {num_procs}',
                            '-r {procs_per_node}', '{extra_args}']

    def get_mpi_specs(self, task, num_procs, num_nodes,
                      procs_per_node, machinefile,
                      hyperthreads, extra_args,
                      resources, workerID):

        # Return auto_resource variables inc. extra_args additions
        if extra_args:
            num_procs, num_nodes, procs_per_node, p_args = \
                self._parse_extra_args(num_procs, num_nodes, procs_per_node,
                                       hyperthreads, extra_args=extra_args)

        rm_rpn = True if procs_per_node is None and num_nodes is None else False

        hostlist = None
        if machinefile and not self.mfile_support:
            logger.warning('User machinefile ignored - not supported by {}'.format(self.run_command))
            machinefile = None
        if machinefile is None and resources is not None:
            num_procs, num_nodes, procs_per_node = \
                mpi_resources.get_resources(resources, num_procs, num_nodes,
                                            procs_per_node, hyperthreads)

            # TODO: Create ERF file if mapping worker to resources req.
        else:
            num_procs, num_nodes, procs_per_node = \
                mpi_resources.task_partition(num_procs, num_nodes,
                                             procs_per_node, machinefile)

        # Remove portable variable if in extra_args
        if extra_args:
            num_procs, num_nodes, procs_per_node = \
                self._rm_replicated_args(num_procs, num_nodes,
                                         procs_per_node, p_args)

        if rm_rpn:
            procs_per_node = None

        return {'num_procs': num_procs,
                'num_nodes': num_nodes,
                'procs_per_node': procs_per_node,
                'extra_args': extra_args,
                'machinefile': machinefile,
                'hostlist': hostlist}



"""
Runs libEnsemble with APOSMM+POUNDERS on the chwirut least squares problem.
All 214 residual calculations for a given point are performed as a single
simulation evaluation.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_pounders.py
   python3 test_persistent_aposmm_pounders.py --nworkers 3 --comms local
   python3 test_persistent_aposmm_pounders.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi tcp
# TESTSUITE_NPROCS: 4
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from math import gamma, pi, sqrt, ceil
from libensemble.sim_funcs.chwirut1 import chwirut_eval as sim_f

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = 'petsc'
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f

from libensemble.gen_funcs.sampling import lhs_sample
from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

# Declare the run parameters/functions
m = 214
n = 3
budget = 10

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('fvec', float, m)],
    'user': {
        'combine_component_func': lambda x: np.sum(np.power(x, 2)),
    },
}

gen_out = [
    ('x', float, n),
    ('x_on_cube', float, n),
    ('sim_id', int),
    ('local_min', bool),
    ('local_pt', bool),
]

# lb tries to avoid x[1]=-x[2], which results in division by zero in chwirut.
lb = (-2 - np.pi / 10) * np.ones(n)
ub = 2 * np.ones(n)

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'fvec'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'localopt_method': 'pounders',
        'rk_const': 0.5 * ((gamma(1 + (n / 2)) * 5) ** (1 / n)) / sqrt(pi),
        'grtol': 1e-6,
        'gatol': 1e-6,
        'dist_to_bound_multiple': 0.5,
        'lhs_divisions': 100,
        'components': m,
        'lb': lb,
        'ub': ub,
    },
}

alloc_specs = {'alloc_f': alloc_f, 'user': {'batch_mode': True, 'num_active_gens': 1}}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 500}

sample_points = np.zeros((0, n))
rand_stream = np.random.RandomState(0)
for i in range(ceil(exit_criteria['sim_max'] / gen_specs['user']['lhs_divisions'])):
    sample_points = np.append(sample_points, lhs_sample(n, gen_specs['user']['lhs_divisions'], rand_stream), axis=0)

gen_specs['user']['sample_points'] = sample_points * (ub - lb) + lb

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert flag == 0
    assert len(H) >= budget

    save_libE_output(H, persis_info, __file__, nworkers)
    # # Calculating the Jacobian at the best point (though this information was not used by pounders)
    # from libensemble.sim_funcs.chwirut1 import EvaluateJacobian
    # J = EvaluateJacobian(H['x'][np.argmin(H['f'])])
    # assert np.linalg.norm(J) < 2000



import numpy as np
import matplotlib.pyplot as plt
from libensemble.libE import libE
from libensemble.tools import add_unique_random_streams
from tutorial_gen import gen_random_sample
from tutorial_sim import sim_find_sine
from mpi4py import MPI

libE_specs = {'comms': 'mpi'}                   # 'nworkers' removed, 'comms' now 'mpi'

nworkers = MPI.COMM_WORLD.Get_size() - 1        # one process belongs to manager
is_manager = (MPI.COMM_WORLD.Get_rank() == 0)   # manager process has MPI rank 0

gen_specs = {'gen_f': gen_random_sample,        # Our generator function
             'out': [('x', float, (1,))],       # gen_f output (name, type, size).
             'user': {'lower': np.array([-3]),  # random sampling lower bound
                      'upper': np.array([3]),   # random sampling upper bound
                      'gen_batch_size': 5       # number of values gen_f will generate per call
                      }
             }

sim_specs = {'sim_f': sim_find_sine,            # Our simulator function
             'in': ['x'],                       # Input field names. 'x' from gen_f output
             'out': [('y', float)]}             # sim_f output. 'y' = sine('x')

persis_info = add_unique_random_streams({}, nworkers+1)  # Initialize manager/workers random streams

exit_criteria = {'sim_max': 80}                 # Stop libEnsemble after 80 simulations

H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info,
                            libE_specs=libE_specs)

# Some (optional) statements to visualize our History array
# Only the manager process should execute this
if is_manager:
    print([i for i in H.dtype.fields])
    print(H)

    colors = ['b', 'g', 'r', 'y', 'm', 'c', 'k', 'w']

    for i in range(1, nworkers + 1):
        worker_xy = np.extract(H['sim_worker'] == i, H)
        x = [entry.tolist()[0] for entry in worker_xy['x']]
        y = [entry for entry in worker_xy['y']]
        plt.scatter(x, y, label='Worker {}'.format(i), c=colors[i-1])

    plt.title('Sine calculations for a uniformly sampled random distribution')
    plt.xlabel('x')
    plt.ylabel('sine(x)')
    plt.legend(loc='lower right')
    plt.savefig('tutorial_sines.png')



import numpy as np
from libensemble.message_numbers import EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources
from libensemble.gen_funcs.old_aposmm import initialize_APOSMM, decide_where_to_start_localopt, update_history_dist


def start_persistent_local_opt_gens(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function will do the following:

    - Start up a persistent generator that is a local opt run at the first point
      identified by APOSMM's decide_where_to_start_localopt. Note, it will do
      this only if at least one worker will be left to perform simulation evaluations.
    - If multiple starting points are available, the one with smallest function
      value is chosen.
    - If no candidate starting points exist, points from existing runs will be
      evaluated (oldest first).
    - If no points are left, call the generation function.

    .. seealso::
        `test_uniform_sampling_then_persistent_localopt_runs.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_then_persistent_localopt_runs.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    Work = {}
    gen_count = support.count_persis_gens()
    points_to_evaluate = ~H['given'] & ~H['cancel_requested']

    # If a persistent localopt run has just finished, use run_order to update H
    # and then remove other information from persis_info
    for i in persis_info.keys():
        if 'done' in persis_info[i]:
            H['num_active_runs'][persis_info[i]['run_order']] -= 1
        if 'x_opt' in persis_info[i]:
            opt_ind = np.all(H['x'] == persis_info[i]['x_opt'], axis=1)
            assert sum(opt_ind) == 1, "There must be just one optimum"
            H['local_min'][opt_ind] = True
            persis_info[i] = {'rand_stream': persis_info[i]['rand_stream']}

    # If wid is idle, but in persistent mode, and its calculated values have
    # returned, give them back to i. Otherwise, give nothing to wid
    for wid in support.avail_worker_ids(persistent=EVAL_GEN_TAG):
        gen_inds = (H['gen_worker'] == wid)
        if support.all_returned(H, gen_inds):
            last_time_pos = np.argmax(H['given_time'][gen_inds])
            last_ind = np.nonzero(gen_inds)[0][last_time_pos]
            Work[wid] = support.gen_work(wid, gen_specs['persis_in'], last_ind, persis_info[wid], persistent=True)
            persis_info[wid]['run_order'].append(last_ind)

    for wid in support.avail_worker_ids(persistent=False):
        # Find candidates to start local opt runs if a sample has been evaluated
        if np.any(np.logical_and(~H['local_pt'], H['returned'], ~H['cancel_requested'])):
            n, _, _, _, r_k, mu, nu = initialize_APOSMM(H, gen_specs)
            update_history_dist(H, n, gen_specs['user'], c_flag=False)
            starting_inds = decide_where_to_start_localopt(H, r_k, mu, nu)
        else:
            starting_inds = []

        # Start persistent generator for local opt run unless it would use all workers
        if starting_inds and gen_count + 1 < len(W):
            # Start at the best possible starting point
            ind = starting_inds[np.argmin(H['f'][starting_inds])]
            try:
                Work[wid] = support.gen_work(wid, gen_specs['persis_in'], ind, persis_info[wid], persistent=True)
            except InsufficientFreeResources:
                break
            H['started_run'][ind] = 1
            H['num_active_runs'][ind] += 1
            persis_info[wid]['run_order'] = [ind]
            gen_count += 1

        elif np.any(points_to_evaluate):

            # Perform sim evaluations from existing runs
            q_inds_logical = np.logical_and(points_to_evaluate, H['local_pt'])
            if not np.any(q_inds_logical):
                q_inds_logical = points_to_evaluate
            sim_ids_to_send = np.nonzero(q_inds_logical)[0][0]  # oldest point
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], sim_ids_to_send, [])
            except InsufficientFreeResources:
                break
            points_to_evaluate[sim_ids_to_send] = False

        elif (gen_count == 0 and not np.any(np.logical_and(W['active'] == EVAL_GEN_TAG, W['persis_state'] == 0))):
            # Finally, generate points since there is nothing else to do (no resource sets req.)
            Work[wid] = support.gen_work(wid, gen_specs.get('in', []), [], persis_info[wid], rset_team=[])
            gen_count += 1

    return Work, persis_info



"""
Runs libEnsemble in order to test the ability of an allocation function to
cancel long-running simulations. In this case, the simulation has a run-time
in seconds that is drawn uniformly from [0,10] and any time the allocation
function is called and a sim_id has been evaluated for more than 5 seconds,
it is cancelled.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_cancel_in_alloc.py
   python3 test_cancel_in_alloc.py --nworkers 3 --comms local
   python3 test_cancel_in_alloc.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.branin.branin_obj import call_branin as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

sim_specs = {
    "sim_f": sim_f,
    "in": ["x"],
    "out": [("f", float)],
    "user": {"uniform_random_pause_ub": 10},
}

gen_specs = {
    "gen_f": gen_f,
    "in": ["sim_id"],
    "out": [("x", float, (2,))],
    "user": {
        "gen_batch_size": 5,
        "lb": np.array([-3, -2]),
        "ub": np.array([3, 2]),
    },
}

alloc_specs = {
    "alloc_f": give_sim_work_first,
    "out": [],
    "user": {
        "cancel_sims_time": 3,
        "batch_mode": False,
        "num_active_gens": 1,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {"sim_max": 10, "elapsed_wallclock_time": 300}

# Perform the run
H, persis_info, flag = libE(
    sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs, alloc_specs=alloc_specs
)

if is_manager:
    test = np.any(H["cancel_requested"]) and np.any(H["kill_sent"])
    assert test, "This test should have requested a cancellation and had a kill sent"
    save_libE_output(H, persis_info, __file__, nworkers)



from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def give_pregenerated_sim_work(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function gives (in order) entries in alloc_spec['x'] to
    idle workers. It is an example use case where no gen_func is used.

    .. seealso::
        `test_fast_alloc.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_fast_alloc.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    Work = {}

    # Unless already defined, initialize next_to_give to be the first point in H
    persis_info['next_to_give'] = persis_info.get('next_to_give', 0)

    if persis_info['next_to_give'] >= len(H):
        return Work, persis_info, 1

    for i in support.avail_worker_ids():
        # Skip any cancelled points
        while persis_info['next_to_give'] < len(H) and H[persis_info['next_to_give']]['cancel_requested']:
            persis_info['next_to_give'] += 1

        # Give sim work
        try:
            Work[i] = support.sim_work(i, H, sim_specs['in'], [persis_info['next_to_give']], [])
        except InsufficientFreeResources:
            break
        persis_info['next_to_give'] += 1

        if persis_info['next_to_give'] >= len(H):
            break

    return Work, persis_info
